﻿using MVR.FileManagementSecure;
using MeshVR;
using SimpleJSON;
using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using System.Linq;
using System.Globalization;

namespace JayJayWon
{
    public class VARPackageManifest
    {
        public static Dictionary<string, VARPackageManifestEntry> varPackages = new Dictionary<string, VARPackageManifestEntry>();

        public static Dictionary<string, VARPackageVersionGroupEntry> varPackageVersionGroups = new Dictionary<string, VARPackageVersionGroupEntry>();

        public static void Update()
        {
            if (VARPackageVersionGroupEntry.anyPersistVARMetaDataPending || VARPackageVersionGroupEntry.anyPersistVARUserDataPending)
                foreach (var vpvg in varPackageVersionGroups) vpvg.Value.Update();

        }

        public static void Init(DateTime scanTime)
        {
            string[] varMetaFiles = FileManagerSecure.GetFiles(Const.VARMetaDataFolder, "*" + Const.MetaDataFileExt);
            string[] varUserFiles = FileManagerSecure.GetFiles(Const.VARUserDataFolder, "*" + Const.UserDataFileExt);

            HashSet<string> varFiles = Utils.GetFilesAtPathRecursive("AddonPackages", "*.var",true);

            HashSet<string> varFileNames = new HashSet<string>() ;
            foreach (string name in varFiles) varFileNames.Add(FileManagerSecure.GetFileName(name));

            ReloadManifest(varMetaFiles, varUserFiles, varFileNames, scanTime);
            RefreshManifest(varFiles, scanTime);
            ReloadVARUserData(varUserFiles);
            RefreshVARUserFilePrefs();
        }

        public static string GetVPVGEKey(string creatorName, string packgageName)
        {
            return creatorName + "." + packgageName;
        }
        public static void ReloadManifest(string[] varMetaFiles, string[] varUserFiles, HashSet<string> varFileNames, DateTime scanTime)
        {
            foreach (string varMetaFile in varMetaFiles)
            {
                try
                {
                    JSONClass varMetaFileJC = SuperController.singleton.LoadJSON(varMetaFile).AsObject;

                    if (varMetaFileJC["VARManifestStoreFormat"].AsInt == Const.VARManifestStoreFormat)
                    {                        
                        JSONArray versionsJA = varMetaFileJC["varVersions"].AsArray;

                        string creatorName = varMetaFileJC["creatorName"];
                        string packageName = varMetaFileJC["packageName"];
                        string vpvgeKey = GetVPVGEKey(creatorName, packageName);

                        bool continueToNextMetaFile = false;
                        foreach (JSONClass jc in versionsJA)
                        {
                            int version = jc["varVersion"].AsInt;
                            string varFileNameFromMetaFile = vpvgeKey + "." + version.ToString()+".var";
                            if (!varFileNames.Contains(varFileNameFromMetaFile))
                            {
                                continueToNextMetaFile = true;
                                break;
                            }
                        }

                        if (continueToNextMetaFile) continue;

                        JSONArray resourcesJA = varMetaFileJC["resources"].AsArray;

                        if (varUserFiles.Contains(Const.VARUserDataFolder + "\\" + creatorName + "." + packageName + Const.UserDataFileExt))
                        {
                            VARPackageVersionGroupEntry vpvge;
                            if (!varPackageVersionGroups.ContainsKey(vpvgeKey))
                            {
                                vpvge = new VARPackageVersionGroupEntry(creatorName, packageName, versionsJA, scanTime);
                                varPackageVersionGroups.Add(vpvgeKey, vpvge);
                                CreatorManifest.RegisterCreator(vpvge.creatorName);
                            }
                            else vpvge = varPackageVersionGroups[vpvgeKey];

                            if (resourcesJA != null) ResourceManifest.RegisterVARResourcesFromManifest(vpvge, resourcesJA);
                        }

                    }
                }
                catch (Exception e) { SuperController.LogError("Exception in BA.ReloadManifest for "+varMetaFile+": " + e); }
            }
        }

        public static void RefreshManifest(HashSet<string> varFiles, DateTime scanTime)
        {            
            foreach (string varFile in varFiles)
            {
                try
                {
                    if (!varPackages.ContainsKey(varFile) && FileManagerSecure.DirectoryExists(varFile + ":"))
                    {
                        VARPackageManifestEntry vpme = new VARPackageManifestEntry(varFile);
                        string vpvgeKey = GetVPVGEKey(vpme.creatorName, vpme.packageName);
                        

                        if (!varPackageVersionGroups.ContainsKey(vpvgeKey) || !varPackageVersionGroups[vpvgeKey].varVersionLicenses.Keys.Contains(vpme.varVersion))
                        {
                            ProcessNewVARFile(vpme, scanTime);
                        }
                        varPackages.Add(varFile, vpme);
                    }
                }
                catch (Exception e) { SuperController.LogError("Exception in BA.RefreshManifest for " + varFile + ": " + e); }               
            }
        }

        public static void ReloadVARUserData(string[] files)
        {
            foreach (string file in files)
            {
                try
                {
                    JSONClass jc = SuperController.singleton.LoadJSON(file).AsObject;

                    string creatorName = jc["creatorName"];
                    string packageName = jc["packageName"];

                    string vpvgeKey = GetVPVGEKey(creatorName, packageName);

                    if (varPackageVersionGroups.ContainsKey(vpvgeKey))
                    {
                        varPackageVersionGroups[vpvgeKey].ReloadVARUserData(jc);
                    }
                }
                catch (Exception e) { SuperController.LogError("Exception in BA.ReloadVARUserData for " + file + ": " + e); }

            }
        }

        public static void RefreshVARUserFilePrefs()
        {
            ResourceManifest.ClearAllUserFilePrefs();
            foreach(var vpme in varPackages)
            {
                if (FileManagerSecure.DirectoryExists("AddonPackagesFilePrefs\\" + vpme.Value.varFileNameNoExtension))
                {
                    string searchFolder = "AddonPackagesFilePrefs\\" + vpme.Value.varFileNameNoExtension;
                    HashSet<string> varFilePrefs = Utils.GetFilesAtPathRecursive(searchFolder, "*.*");
                    foreach (string varFilePref in varFilePrefs)
                    {
                        try
                        {
                            int fileExtIndex = varFilePref.LastIndexOf('.');
                            int filePathLength = varFilePref.Length;

                            string rvgeKey = ResourceManifest.GetRVGKey(vpme.Value, varFilePref.Substring(searchFolder.Length + 1, fileExtIndex - searchFolder.Length - 1));
                            if (varFilePref.EndsWith(".hide")) ResourceManifest.UpdateVARUserHideFlag(rvgeKey, vpme.Value.varVersion, true);
                            if (varFilePref.EndsWith(".fav")) ResourceManifest.UpdateVARUserFavFlag(rvgeKey, vpme.Value.varVersion, true);
                        }
                        catch (Exception e) { SuperController.LogError("Exception in BA.RefreshVARUserFilePrefs for VAR " + vpme.Value.varFileNameNoExtension + "and file "+ varFilePref+": " + e); }

                    }
                }

            }
        }

        private static void ProcessNewVARFile(VARPackageManifestEntry vpme, DateTime scanTime)
        {

            string vpvgeKey = GetVPVGEKey(vpme.creatorName, vpme.packageName);
            VARPackageVersionGroupEntry vpvge;
            if (!varPackageVersionGroups.ContainsKey(vpvgeKey))
            {
                vpvge = new VARPackageVersionGroupEntry(vpme,scanTime);
                varPackageVersionGroups.Add(vpvgeKey, vpvge);
                CreatorManifest.RegisterCreator(vpvge.creatorName);                
            }
            else
            {
                vpvge = varPackageVersionGroups[vpvgeKey];
                if (vpvge.varVersionLicenses.Keys.Contains(vpme.varVersion)) return;
                vpvge.AddVARVersion(vpme.varVersion,vpme.varFilePath);
            }

            vpvge.SetPersistVARMetaData();
            vpvge.SetPersistVARUserData();
            vpme.vpvge = vpvge;
            ResourceManifest.RegisterVARResources(vpme);

        }
    }

    public class VARPackageManifestEntry
    {
        public string varFilePath { get; private set; }
        public string varFileNameNoExtension { get; private set; }
        public string creatorName { get; private set; }
        public string packageName { get; private set; }
        public int varVersion { get; private set; }

        public VARPackageVersionGroupEntry vpvge;

        public VARPackageManifestEntry(string varFile)
        {
            varFilePath = varFile;
            string varFileName = FileManagerSecure.GetFileName(varFilePath);
            varFileNameNoExtension = varFileName.Substring(0, varFileName.Length - 4);

            string[] parts = varFileNameNoExtension.Split('.');
            creatorName = parts[0];
            packageName = parts[1];
            varVersion = int.Parse(parts[2]);    
        }
    }
    public class VARPackageVersionGroupEntry
    {
        public string creatorName { get; private set; }
        public string packageName { get; private set; }

        private HashSet<int> varVersions = new HashSet<int>();
        public Dictionary<int,int> varVersionLicenses = new Dictionary<int, int>();

        private Dictionary<int, DateTime> varVersionFirstScanTimes = new Dictionary<int, DateTime>();
        private Dictionary<int, DateTime> varVersionFirstCacheTimes = new Dictionary<int, DateTime>();

        public DateTime firstCataloguedDateTime
        {
            get
            {
                DateTime firstScanDateTime = varVersionFirstScanTimes.Values.Min(d => d);
                if (varVersionFirstCacheTimes.Count == 0) return firstScanDateTime;
                DateTime firstCacheDateTime = varVersionFirstCacheTimes.Values.Min(d => d);
                return new DateTime(Math.Min(firstScanDateTime.Ticks, firstCacheDateTime.Ticks));
            }
        }

        public int highestVARVersion { get; private set; }

        private HashSet<PackagedRVGE> packageRVGEs = new HashSet<PackagedRVGE>();
        private Dictionary<string,PackagedRVGE> packageRVGEDict = new Dictionary<string, PackagedRVGE>();

        private bool persistVARMetaDataPending = false;
        private bool persistVARUserDataPending = false;
        public static bool anyPersistVARMetaDataPending { get; private set; } = false;
        public static bool anyPersistVARUserDataPending { get; private set; } = false;

        public VARPackageVersionGroupEntry(VARPackageManifestEntry vpme, DateTime scanTime)
        {
            creatorName = vpme.creatorName;
            packageName = vpme.packageName;

            varVersions.Add(vpme.varVersion);
            varVersionLicenses.Add(vpme.varVersion, GetLicenseTypeFromMeta(vpme.varFilePath));
            varVersionFirstScanTimes.Add(vpme.varVersion, scanTime);
            highestVARVersion = vpme.varVersion;

        }

        public VARPackageVersionGroupEntry(string _creatorName, string _packageName, JSONArray varVersionsJA, DateTime scanTime)
        {
            creatorName = _creatorName;
            packageName = _packageName;
            highestVARVersion = 0;
            foreach (JSONClass jc in varVersionsJA)
            {
                int varVersion = jc["varVersion"].AsInt;
                varVersions.Add(varVersion);
                varVersionLicenses.Add(varVersion, EnumManifest.GetEnumValFromStoreVal(LicenseType.enumManifestName,jc["varLicense"]));
                varVersionFirstScanTimes.Add(varVersion, scanTime);

                if (highestVARVersion < varVersion) highestVARVersion = varVersion;
            }
            
        }

        private int GetLicenseTypeFromMeta(string varFileName)
        {
            int varLicenseType = LicenseType.Unknown;
            if (FileManagerSecure.FileExists(varFileName + ":\\meta.json"))
            {
               JSONClass varMetaJSON = SuperController.singleton.LoadJSON(varFileName + ":\\meta.json").AsObject;
                if (varMetaJSON != null && varMetaJSON["licenseType"] != null)
                {
                    string varLicenseTypeStoreVal = varMetaJSON["licenseType"];
                    if (EnumManifest.GetStoreValEnumsDict(LicenseType.enumManifestName).Keys.Contains(varLicenseTypeStoreVal))
                    {
                        varLicenseType = EnumManifest.GetEnumValFromStoreVal(LicenseType.enumManifestName, varLicenseTypeStoreVal);
                    }
                    else
                    {
                        SuperController.LogMessage(varFileName + " contains unrecognised License Type. Possibly caused by creator manually editing VAR meta.json");
                    }
                    
                }
            }
            else SuperController.LogMessage("Meta Not found: "+ varFileName);
            return varLicenseType;
        }
        public void SetPersistVARMetaData()
        {
            persistVARMetaDataPending = true;
            anyPersistVARMetaDataPending = true;
        }

        public void SetPersistVARUserData()
        {
            persistVARUserDataPending = true;
            anyPersistVARUserDataPending = true;
        }

        public void AddVARVersion(int varVersion, string varFilePath)
        {
            varVersions.Add(varVersion);
            varVersionLicenses.Add(varVersion, GetLicenseTypeFromMeta(varFilePath));
            if (varVersion> highestVARVersion) highestVARVersion = varVersion;
        }

        public void RegisterRVGE(PackagedRVGE packagedRVGE)
        {
            packageRVGEs.Add(packagedRVGE);
            packageRVGEDict.Add(ResourceManifest.GetRVGKey(this, packagedRVGE.resourceFullFileName),packagedRVGE);
        }

        public void ReloadVARUserData(JSONClass varUserDataJC)
        {
            int varUserDataStoreFormat = varUserDataJC["VARUserDataStoreFormat"].AsInt;
            JSONArray varVersionsJA = varUserDataJC["varVersions"].AsArray;
            JSONArray resourcesJA = varUserDataJC["resources"].AsArray;

            HashSet<int> previouslyScannedVARVersions = new HashSet<int>();

            if (varUserDataStoreFormat == 1)
            {
                foreach (JSONNode varVersionJC in varVersionsJA)
                {
                    int varVersion = varVersionJC.AsInt;
                    previouslyScannedVARVersions.Add(varVersion);
                }
                foreach (JSONClass resourceJC in resourcesJA)
                {
                    JSONArray resourceVARVersionsJA = resourceJC["varVersions"].AsArray;
                    foreach (JSONClass resourceVARVersionJC in resourceVARVersionsJA)
                    {
                        int varVersion = resourceVARVersionJC["varVersion"].AsInt;
                        DateTime firstScanTime = DateTime.Parse(resourceVARVersionJC["firstScanDateTime"], CultureInfo.InvariantCulture, DateTimeStyles.RoundtripKind);

                        if (varVersionFirstScanTimes.ContainsKey(varVersion)) varVersionFirstScanTimes[varVersion] = firstScanTime;
                        else varVersionFirstScanTimes.Add(varVersion, firstScanTime);
                    }
                }
                SetPersistVARUserData();
            }
            else
            {
                foreach (JSONClass varVersionJC in varVersionsJA)
                {
                    int varVersion = varVersionJC["varVersion"].AsInt;
                    previouslyScannedVARVersions.Add(varVersion);

                    DateTime firstScanTime = DateTime.Parse(varVersionJC["firstScanDateTime"], CultureInfo.InvariantCulture, DateTimeStyles.RoundtripKind);
                    if (varVersionFirstScanTimes.ContainsKey(varVersion)) varVersionFirstScanTimes[varVersion] = firstScanTime;
                    else varVersionFirstScanTimes.Add(varVersion, firstScanTime);

                    DateTime? firstCachedDateTime =null;
                    string cacheMetaFileName = "Cache\\PackageJSON\\" + creatorName + "." + packageName + "." + varVersion.ToString() + ".vamcachemeta";
                    if (varVersionJC["firstCacheDateTime"] != null)
                    {
                        firstCachedDateTime = DateTime.Parse(varVersionJC["firstCacheDateTime"], CultureInfo.InvariantCulture, DateTimeStyles.RoundtripKind);
                        
                    }
                    else if (FileManagerSecure.FileExists(cacheMetaFileName))
                    {
                        JSONClass cacheMetaFileJC = SuperController.singleton.LoadJSON(cacheMetaFileName).AsObject;
                        if (cacheMetaFileJC["date"]!=null) firstCachedDateTime = DateTime.Parse(cacheMetaFileJC["date"], new CultureInfo("en-US", false));
                        SetPersistVARUserData();
                    }
                    if (firstCachedDateTime != null)
                    {
                        if (varVersionFirstCacheTimes.ContainsKey(varVersion)) varVersionFirstCacheTimes[varVersion] = (DateTime)firstCachedDateTime;
                        else varVersionFirstCacheTimes.Add(varVersion, (DateTime)firstCachedDateTime);
                    }
                }              
            }

            foreach (JSONClass resourceJC in resourcesJA)
            {
                string rvgeKey = ResourceManifest.GetRVGKey(creatorName, packageName, resourceJC["resourceFullFileName"]);
                if (packageRVGEDict.ContainsKey(rvgeKey))
                {
                    ResourceVersionGroupEntry rvge = packageRVGEDict[rvgeKey];
                    if (!rvge.localFileResource)
                    {
                        (rvge as PackagedRVGE).RestoreVARUserData(resourceJC);
                    }
                }
            }

            HashSet<int> newVARVersions = new HashSet<int>(varVersions.Except(previouslyScannedVARVersions));
            if (newVARVersions.Count > 0)
            {
                foreach (ResourceVersionGroupEntry rvge in packageRVGEs)
                {
                    if (!rvge.localFileResource) (rvge as PackagedRVGE).NewVARVersions(newVARVersions);
                }
                SetPersistVARUserData();
            }

        }

        public void Update()
        {
            if (persistVARMetaDataPending )
            {
                JSONClass jc = new JSONClass();
                jc["VARManifestStoreFormat"].AsInt = Const.VARManifestStoreFormat;

                jc["BAMajorVersion"].AsInt = Const.BAMajorVersion;
                jc["BAMinorVersion"].AsInt = Const.BAMinorVersion;
                jc["BAFixVersion"].AsInt = Const.BAFixVersion;

                jc["creatorName"] = creatorName;
                jc["packageName"] = packageName;

                JSONArray varVersionsJA = new JSONArray();
                jc["varVersions"] = varVersionsJA;
                foreach (var varVersion in varVersionLicenses)
                {
                    JSONClass varVersionJC = new JSONClass();
                    varVersionJC["varVersion"].AsInt = varVersion.Key;
                    varVersionJC["varLicense"] = EnumManifest.GetStoreVal(LicenseType.enumManifestName, varVersion.Value);
                    varVersionsJA[-1] = varVersionJC;
                }             

                JSONArray ja = new JSONArray();
                foreach(ResourceVersionGroupEntry rvge in packageRVGEs)
                {
                    ja[-1] = rvge.GetCoreJSON();
                }

                jc["resources"] = ja;
                SuperController.singleton.SaveJSON(jc, Const.VARMetaDataFolder + "\\" + creatorName + "." + packageName+Const.MetaDataFileExt);
                persistVARMetaDataPending = false;
            }

            if (persistVARUserDataPending)
            {
                JSONClass varJC = new JSONClass();
                varJC["VARUserDataStoreFormat"].AsInt = Const.VARUserDataStoreFormat;

                varJC["BAMajorVersion"].AsInt = Const.BAMajorVersion;
                varJC["BAMinorVersion"].AsInt = Const.BAMinorVersion;
                varJC["BAFixVersion"].AsInt = Const.BAFixVersion;

                varJC["creatorName"] = creatorName;
                varJC["packageName"] = packageName;

                JSONArray varVersionsJA = new JSONArray();
                varJC["varVersions"] = varVersionsJA;
                foreach (var varVersionFirstScanTime in varVersionFirstScanTimes)
                {
                    JSONClass varVersionJC = new JSONClass();
                    varVersionJC["varVersion"].AsInt = varVersionFirstScanTime.Key;
                    varVersionJC["firstScanDateTime"] = varVersionFirstScanTime.Value.ToString("O");
                    if (varVersionFirstCacheTimes.ContainsKey(varVersionFirstScanTime.Key))
                    {
                        varVersionJC["firstCacheDateTime"] = varVersionFirstCacheTimes[varVersionFirstScanTime.Key].ToString("O");
                    }

                    varVersionsJA[-1] = varVersionJC;
                }

                JSONArray resourcesJA = new JSONArray();
                varJC["resources"] = resourcesJA;
                foreach (var rvge in packageRVGEs)
                {
                    resourcesJA[-1] = rvge.GetVARUserDataJSON();
                }

                SuperController.singleton.SaveJSON(varJC, Const.VARUserDataFolder + "\\" + creatorName + "." + packageName + Const.UserDataFileExt);
                persistVARUserDataPending = false;
            }

            anyPersistVARMetaDataPending = false;
            anyPersistVARUserDataPending = false;
        }
    }

    public class ResourceManifest
    {
        private static Dictionary<string, ResourceVersionGroupEntry> resourceVersionGroupsDict = new Dictionary<string, ResourceVersionGroupEntry>();

        private static HashSet<ResourceVersionGroupEntry> resourceVersionGroups = new HashSet<ResourceVersionGroupEntry>();

        private static HashSet<LocalRVGE> localResourceVersionGroups = new HashSet<LocalRVGE>();
        private static HashSet<PackagedRVGE> varResourceVersionGroups = new HashSet<PackagedRVGE>();

        public static void Update()
        {
            if (LocalRVGE.anyPersistLocalRVGE)
            {
                foreach (var rvge in localResourceVersionGroups) rvge.SaveLocalUserData();
            }
        }
        public static void LocalResourcesInit(DateTime scanTime)
        {
            foreach (LocalRVGE existingLocalRVGE in localResourceVersionGroups)
            {
                resourceVersionGroups.Remove(existingLocalRVGE);
                resourceVersionGroupsDict.Remove(existingLocalRVGE.resourceFullFileName);
            }
            localResourceVersionGroups.Clear();

            RefreshLocalSceneManifest(scanTime);
            RefreshLocalPresetManifest("Custom\\Atom\\Person\\Appearance", ResourceType.presetAppearance, scanTime);
            RefreshLocalPresetManifest("Custom\\Atom\\Person\\AnimationPresets", ResourceType.presetAnimation, scanTime);
            RefreshLocalPresetManifest("Custom\\Atom\\Person\\BreastPhysics", ResourceType.presetBreastPhys, scanTime);
            RefreshLocalPresetManifest("Custom\\Atom\\Person\\Clothing", ResourceType.presetClothing, scanTime);
            RefreshLocalPresetManifest("Custom\\Atom\\Person\\GlutePhysics", ResourceType.presetGlutePhysics, scanTime);
            RefreshLocalPresetManifest("Custom\\Atom\\Person\\Hair", ResourceType.presetHair, scanTime);
            RefreshLocalPresetManifest("Custom\\Atom\\Person\\Morphs", ResourceType.presetMorphs, scanTime);
            RefreshLocalPresetManifest("Custom\\Atom\\Person\\Plugins", ResourceType.presetPlugins, scanTime);
            RefreshLocalPresetManifest("Custom\\Atom\\Person\\Pose", ResourceType.presetPose, scanTime);
            RefreshLocalPresetManifest("Custom\\Atom\\Person\\Skin", ResourceType.presetSkin, scanTime);

            string[] patterns = new string[3] {"*.cs","*.cslist","*.dll" };
            HashSet<string> localPluginFiles = Utils.GetFilesAtPathRecursive("Custom\\Scripts", patterns);
            RefreshLocalResourceManifest(localPluginFiles, ResourceType.plugins, scanTime);
        }
        private static void RefreshLocalSceneManifest(DateTime scanTime)
        {
            HashSet<string> localFiles = Utils.GetFilesAtPathRecursive("Saves\\scene", "*.json");

            RefreshLocalResourceManifest(localFiles, ResourceType.scene, scanTime);
        }
        private static void RefreshLocalPresetManifest(string path, int resourceType, DateTime scanTime)
        {
            HashSet<string> localFiles = Utils.GetFilesAtPathRecursive(path, "*.vap");

            RefreshLocalResourceManifest(localFiles, resourceType, scanTime);
        }

        private static void RefreshLocalResourceManifest(HashSet<string> localFiles, int resourceType, DateTime scanTime)
        {   

            foreach (string localFile in localFiles)
            {
                try
                {
                    FileManagerSecure.CreateDirectory(Const.LocalUserDataFolder + "\\" + FileManagerSecure.GetDirectoryName(localFile));

                    RegisterLocalResourceOfType(localFile, resourceType, scanTime);
                }
                catch (Exception e) { SuperController.LogError("Exception in BA.RefreshLocalResourceManifest for " + localFile + ": " + e); }

            }
        }

        public static void UpdateVARUserHideFlag(string rvgKey,int varVersion, bool hidden)
        {
            if (resourceVersionGroupsDict.ContainsKey(rvgKey))
            {
                ResourceVersionGroupEntry rvge = resourceVersionGroupsDict[rvgKey];
                if (!rvge.localFileResource) 
                {
                    (rvge as PackagedRVGE).SetHideUserVARVersion(varVersion, hidden);
                }
                
            }
            
        }
        public static void UpdateVARUserFavFlag(string rvgKey, int varVersion, bool favourite)
        {
            if (resourceVersionGroupsDict.ContainsKey(rvgKey))
            {
                ResourceVersionGroupEntry rvge = resourceVersionGroupsDict[rvgKey];
                if (!rvge.localFileResource)
                {
                    (rvge as PackagedRVGE).SetFavUserVARVersion(varVersion, favourite);
                }
                    
            }

        }

        public static List<ResourceVersionGroupEntry> GetResourceGroupsByCreator(int resourceType, bool showHidden, bool onlyFav, JSONStorableMultiEnumStringChooser tagsFilterJSMultiEnum, string creator)
        {
            var filteredRVGs = resourceVersionGroups.Where(o =>
                        o.resourceType == resourceType && o.resourcePresent && 
                        (!o.isHidden || showHidden) &&
                        (o.isFavourite || !onlyFav) &&
                        !o.isCSLReferenced &&
                        o.creatorName == creator);
            if (tagsFilterJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.mainValFlag)
            {
                string tagFilterName = tagsFilterJSMultiEnum.mainVal;
                return filteredRVGs.Where(o => o.IsTagged(TagManifest.GetTag(resourceType, tagFilterName))).ToList();
            }
            else if (tagsFilterJSMultiEnum.valTopEnum == TagGroups.hiddenOnly) return filteredRVGs.Where(o => o.isHidden).ToList();
            else return filteredRVGs.ToList();
            
        }

        public static List<int> GetLicenseTypesByInverseFilters(int resourceType, bool showHidden, bool onlyFav, JSONStorableMultiEnumStringChooser tagsFilterJSMultiEnum, JSONStorableMultiEnumStringChooser creatorFilterJSMultiEnum)
        {
            List<int> licenseTypesNotPresent = new List<int>();
            bool anyCreator = (creatorFilterJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.topEnumValFlag && creatorFilterJSMultiEnum.valTopEnum == CreatorFilterChoices.allVARs);
            var filteredRVGs = resourceVersionGroups.Where(o =>
                        o.resourceType == resourceType && o.resourcePresent &&
                        (!o.isHidden || showHidden) &&
                        (o.isFavourite || !onlyFav) &&
                        !o.isCSLReferenced &&
                        !o.localFileResource &&
                        (anyCreator || o.creatorName == creatorFilterJSMultiEnum.mainVal));

            IEnumerable<ResourceVersionGroupEntry> tagFilteredRVGEs = filteredRVGs;
            if (tagsFilterJSMultiEnum.valType != JSONStorableMultiEnumStringChooser.topEnumValFlag)
            {
                tagFilteredRVGEs = filteredRVGs.Where(o => o.IsTagged(TagManifest.GetTag(resourceType, tagsFilterJSMultiEnum.mainVal))).ToList();
            }
            else
            {
                if (tagsFilterJSMultiEnum.valTopEnum == TagGroups.hiddenOnly) tagFilteredRVGEs = filteredRVGs.Where(o => o.isHidden).ToList();
                if (tagsFilterJSMultiEnum.valTopEnum == TagGroups.noTags) tagFilteredRVGEs = filteredRVGs.Where(o => o.IsNotTagged()).ToList();
                if (tagsFilterJSMultiEnum.valTopEnum == TagGroups.noUserTags) tagFilteredRVGEs = filteredRVGs.Where(o => o.IsNotUserTagged()).ToList();
            }


            foreach (int licenseType  in EnumManifest.GetEnumsList(LicenseType.enumManifestName))
            {
                if (tagFilteredRVGEs.Where(o => ((PackagedRVGE)o).varLicenseType == licenseType).Count() == 0) licenseTypesNotPresent.Add(licenseType);
            }
            licenseTypesNotPresent.Remove(LicenseType.Any);
            return licenseTypesNotPresent;
        }

        public static List<ResourceVersionGroupEntry> GetResourceGroups(BrowserFiltersUI filters)
        {
            var globalFilterRVGs = resourceVersionGroups.Where(o =>
                        o.resourceType == filters.resourceTypeJSEnum.val && o.resourcePresent &&
                        (!o.isHidden || filters.showHiddenJSB.val) &&
                        (o.isFavourite || !filters.onlyFavouritesJSB.val) &&
                        !o.isCSLReferenced &&
                        (o.resourceDisplayNameLowerCaps.StartsWith(filters.resourceNameFilterQuick) || filters.resourceNameFilterQuick == ""));

            if (filters.sortModeJSEnum.val == BrowserSortMode.AtoZ) globalFilterRVGs = globalFilterRVGs.OrderBy(o => o.resourceDisplayName);
            else if (filters.sortModeJSEnum.val == BrowserSortMode.ZtoA) globalFilterRVGs = globalFilterRVGs.OrderByDescending(o => o.resourceDisplayName);
            else if (filters.sortModeJSEnum.val == BrowserSortMode.cataloguedDateTime)
                 globalFilterRVGs = globalFilterRVGs.OrderByDescending(o => o.firstCataloguedDateTime).ThenBy(o =>o.creatorName+ o.packageName + o.folderName).ThenBy(o => o.resourceDisplayName);
            else if (filters.sortModeJSEnum.val == BrowserSortMode.packageFolder) globalFilterRVGs = globalFilterRVGs.OrderBy(o => o.creatorName + o.packageName + o.folderName).ThenBy(o => o.resourceDisplayName);

            IEnumerable<ResourceVersionGroupEntry> tagFilteredRVGEs = globalFilterRVGs ;
            if (filters.tagsFilterJSMultiEnum.valType != JSONStorableMultiEnumStringChooser.topEnumValFlag)
            {
                tagFilteredRVGEs = globalFilterRVGs.Where(o => o.IsTagged(TagManifest.GetTag(filters.resourceTypeJSEnum.val, filters.tagsFilterJSMultiEnum.mainVal))).ToList();
            }
            else
            {
                if (filters.tagsFilterJSMultiEnum.valTopEnum == TagGroups.hiddenOnly) tagFilteredRVGEs = globalFilterRVGs.Where(o => o.isHidden).ToList();
                if (filters.tagsFilterJSMultiEnum.valTopEnum == TagGroups.noTags) tagFilteredRVGEs = globalFilterRVGs.Where(o => o.IsNotTagged()).ToList();
                if (filters.tagsFilterJSMultiEnum.valTopEnum == TagGroups.noUserTags) tagFilteredRVGEs = globalFilterRVGs.Where(o => o.IsNotUserTagged()).ToList();
            }

            IEnumerable<ResourceVersionGroupEntry> creatorFilteredRVGEs = tagFilteredRVGEs;
            if (filters.creatorJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.topEnumValFlag)
            {
                if (filters.creatorJSMultiEnum.valTopEnum == CreatorFilterChoices.allVARs)
                {
                    creatorFilteredRVGEs = tagFilteredRVGEs.Where(o =>
                        !o.localFileResource).ToList();
                }
                else if (filters.creatorJSMultiEnum.valTopEnum == CreatorFilterChoices.localOnly)
                {
                    creatorFilteredRVGEs = tagFilteredRVGEs.Where(o => o.localFileResource).ToList();
                }
            }
            else
            {
                creatorFilteredRVGEs = tagFilteredRVGEs.Where(o => o.creatorName == filters.creatorJSMultiEnum.displayVal).ToList();
            }
            IEnumerable<ResourceVersionGroupEntry> licenseFilteredRVGEs = creatorFilteredRVGEs;

            if (filters.licenseFilterJSEnum.val!=LicenseType.Any &&( filters.creatorJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.mainValFlag || filters.creatorJSMultiEnum.valTopEnum == CreatorFilterChoices.allVARs))
            {
                licenseFilteredRVGEs = creatorFilteredRVGEs.Where(o => ((PackagedRVGE)o).varLicenseType == filters.licenseFilterJSEnum.val);
            }

            return licenseFilteredRVGEs.ToList();
        }
        public static string GetRVGKey(string creatorName, string packageName, string resourceFileName)
        {
            return creatorName + "." + packageName + ":" + resourceFileName;
        }
        public static string GetRVGKey (VARPackageManifestEntry vpme, string resourceFileName)
        {
            return GetRVGKey(vpme.creatorName, vpme.packageName, resourceFileName);
        }
        public static string GetRVGKey(VARPackageVersionGroupEntry vpvge, string resourceFileName)
        {
            return GetRVGKey(vpvge.creatorName, vpvge.packageName, resourceFileName);
        }

        public static void RegisterVARResources(VARPackageManifestEntry vpme)
        {
            HashSet<string> contentFiles ;
            HashSet<string> hideFiles = new HashSet<string>();
            HashSet<string> favFiles = new HashSet<string>();

            contentFiles = Utils.GetFilesAtPathRecursive(vpme.varFilePath + ":", "*.*");

            int length = vpme.varFilePath.Length + 2;
            foreach (string contentFile in contentFiles)
            {
                if (contentFile.EndsWith(".hide")) hideFiles.Add(contentFile);
                else if (contentFile.EndsWith(".fav")) favFiles.Add(contentFile);
                else
                {
                    int resourceType = GetResourceType(contentFile.Substring(length));
                    if (resourceType!=ResourceType.unknown) RegisterVARResourceOfType(vpme, contentFile.Substring(length), resourceType);
                }
            }

            HashSet<string> csListFiles =  new HashSet<string>(contentFiles.Where(fileName => fileName.EndsWith(".cslist") && fileName.Contains(":\\Custom\\Scripts")));
            ProcessVARCSListReferences(vpme, csListFiles);

            RegisterCreatorHideFlags(vpme, hideFiles);
            RegisterCreatorFavFlags(vpme, favFiles);
        }

        public static void RegisterVARResourcesFromManifest(VARPackageVersionGroupEntry vpvge, JSONArray resourcesJA)
        {
            foreach (JSONClass resourceJC in resourcesJA)
            {
                string resourceFullFileName = resourceJC["resourceFullFileName"];
                string rvgKey = GetRVGKey(vpvge, resourceFullFileName);
                PackagedRVGE rvge = new PackagedRVGE(resourceJC,vpvge);
                resourceVersionGroupsDict.Add(rvgKey, rvge);
                resourceVersionGroups.Add(rvge);
                varResourceVersionGroups.Add(rvge);
                CreatorManifest.RegisterResource(rvge);
                vpvge.RegisterRVGE(rvge);
            }
        }
        private static void RegisterCreatorHideFlags(VARPackageManifestEntry vpme, HashSet<string> hideFiles)
        {
            foreach (string hideFile in hideFiles)
            {               
                string hideFileNoExt = hideFile.Substring(0, hideFile.Length - 5);
                string hideFilPath = hideFileNoExt.Substring(hideFileNoExt.IndexOf(":\\") + 2);

                string rvgKey = GetRVGKey(vpme, hideFilPath);
                
                if (resourceVersionGroupsDict.ContainsKey(rvgKey))
                {
                    ResourceVersionGroupEntry rvge = resourceVersionGroupsDict[rvgKey];
                    (rvge as PackagedRVGE).SetHideCreatorVARVersion(vpme.varVersion);
                }
            }
        }
        private static void RegisterCreatorFavFlags(VARPackageManifestEntry vpme, HashSet<string> favFiles)
        {
            foreach (string favFile in favFiles)
            {
                string favFileNoExt = favFile.Substring(0, favFile.Length - 4);
                string favFilPath = favFileNoExt.Substring(favFileNoExt.IndexOf(":\\") + 2);

                string rvgKey = GetRVGKey(vpme, favFilPath);
                if (resourceVersionGroupsDict.ContainsKey(rvgKey))
                {
                    ResourceVersionGroupEntry rvge = resourceVersionGroupsDict[rvgKey];
                    (rvge as PackagedRVGE).SetFavCreatorVARVersion(vpme.varVersion);
                }
            }
        }

        private static void ProcessVARCSListReferences(VARPackageManifestEntry vpme, HashSet<string> csListFiles)
        {

            string[] stringSeparators = new string[] { ".cs", "\n", "\r\n" };
            string csListFileContents = null;

            foreach (string cslFilePath in csListFiles)
            {
                try
                {
                    csListFileContents = FileManagerSecure.ReadAllText(cslFilePath);
                }
                catch (Exception e)
                {
                    Utils.LogDebug("<ResourceManifest.ProcessCSListReferences> '" + cslFilePath + "' could not be read by VAM - possible malformed file");
                    break;
                }
                if (csListFileContents != null)
                {
                    string[] csListPluginFiles = csListFileContents.Split(stringSeparators, System.StringSplitOptions.RemoveEmptyEntries);

                    string cslPath = FileManagerSecure.GetDirectoryName(cslFilePath.Substring(cslFilePath.IndexOf(":\\")+2));
                    foreach (string csFileRef in csListPluginFiles)
                    {
                        if (!(String.IsNullOrEmpty(csFileRef) || csFileRef.Trim().Length == 0) && !csFileRef.Contains(":"))
                        {
                            string rvgKey = GetRVGKey(vpme, cslPath+"\\"+csFileRef.Replace('/','\\') + ".cs");
                            if (resourceVersionGroupsDict.ContainsKey(rvgKey))
                            {
                                 var rvge = (PackagedRVGE) resourceVersionGroupsDict[rvgKey];
                                rvge.SetCSLFileReferenced(vpme.varVersion);
                            }
                            

                        }

                    }
                }
            }
                
        }

        private static int GetResourceType(string resourceFileName)
        {
            if (resourceFileName.StartsWith("Saves\\scene") && resourceFileName.EndsWith(".json")) return ResourceType.scene;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\AnimationPresets") && resourceFileName.EndsWith(".vap")) return ResourceType.presetAnimation;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Appearance") && resourceFileName.EndsWith(".vap")) return ResourceType.presetAppearance;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\BreastPhysics") && resourceFileName.EndsWith(".vap")) return ResourceType.presetBreastPhys;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Clothing") && resourceFileName.EndsWith(".vap")) return ResourceType.presetClothing;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\General") && resourceFileName.EndsWith(".vap")) return ResourceType.presetGeneral;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\GlutePhysics") && resourceFileName.EndsWith(".vap")) return ResourceType.presetGlutePhysics;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Hair") && resourceFileName.EndsWith(".vap")) return ResourceType.presetHair;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Morphs") && resourceFileName.EndsWith(".vap")) return ResourceType.presetMorphs;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Plugins") && resourceFileName.EndsWith(".vap")) return ResourceType.presetPlugins;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Pose") && resourceFileName.EndsWith(".vap")) return ResourceType.presetPose;
            if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Skin") && resourceFileName.EndsWith(".vap")) return ResourceType.presetSkin;
            //           if (resourceFileName.StartsWith("Custom\\Atom\\Person\\Textures") && (resourceFileName.EndsWith(".jpg")|| resourceFileName.EndsWith(".png"))) return ResourceType.personTexture;
            //           if (resourceFileName.StartsWith("Custom\\Textures") && (resourceFileName.EndsWith(".jpg") || resourceFileName.EndsWith(".png"))) return ResourceType.atomTexture;

            if (resourceFileName.StartsWith("Custom\\Scripts") && (resourceFileName.EndsWith(".cs") || resourceFileName.EndsWith(".cslist") || resourceFileName.EndsWith(".dll"))) return ResourceType.plugins;

            return ResourceType.unknown;
        }

        private static void RegisterVARResourceOfType(VARPackageManifestEntry vpme, string resourceFileName,int resourceType)
        {
            string rvgKey = GetRVGKey(vpme, resourceFileName);
            if (!resourceVersionGroupsDict.ContainsKey(rvgKey))
            {
                PackagedRVGE rvge = new PackagedRVGE(vpme, resourceFileName, resourceType);
                resourceVersionGroupsDict.Add(rvgKey, rvge);
                resourceVersionGroups.Add(rvge);
                varResourceVersionGroups.Add(rvge);
                CreatorManifest.RegisterResource(rvge);
                vpme.vpvge.RegisterRVGE(rvge);
            }
            else
            {
                (resourceVersionGroupsDict[rvgKey] as PackagedRVGE).AddVARVersion(vpme.varVersion);
            }
        }

        private static void RegisterLocalResourceOfType(string resourceFileName, int resourceType, DateTime scanTime)
        {
            if (!resourceVersionGroupsDict.ContainsKey(resourceFileName))
            {
                LocalRVGE rvge = new LocalRVGE(resourceFileName, resourceType);
                resourceVersionGroupsDict.Add(resourceFileName, rvge);
                resourceVersionGroups.Add(rvge);
                localResourceVersionGroups.Add(rvge);

                string localUserDataManifestFilename = Const.LocalUserDataFolder + "\\" + rvge.resourceFullFileName + Const.MetaDataFileExt;
                if (FileManagerSecure.FileExists(localUserDataManifestFilename))
                {
                    rvge.LoadLocalUserData(localUserDataManifestFilename);
                }
                else
                {
                    rvge.SetLocalFileScanTime(scanTime);
                    rvge.SetUserDataPersist();
                }
            }
            else
            {
                SuperController.LogMessage("TBD - Re-register local files");
            }
        }

        public static void ClearAllUserFilePrefs()
        {
            foreach (PackagedRVGE packegedRVGE in varResourceVersionGroups)
            {
                packegedRVGE.ClearAllUserFilePrefs();
            }
        }


    }


    public abstract class ResourceVersionGroupEntry
    {
        protected int baFavourite = TriState.indeterminate;
        protected int baHidden = TriState.indeterminate;
        public virtual string creatorName { get; set; }
        public virtual string packageName { get; set; }
        public virtual bool resourcePresent { get { return true; } }
        public string folderName { get; private set; }
        public string resourceName { get; private set; }
        public string resourceDisplayName { get
            {
                if (resourceCategory == ResourceCategory.preset) return resourceName.Substring(7);
                return resourceName;
            }
        }

        public string resourceNameLowerCaps { get; private set; }
        public string resourceDisplayNameLowerCaps
        {
            get
            {
                if (resourceCategory == ResourceCategory.preset) return resourceNameLowerCaps.Substring(7);
                return resourceNameLowerCaps;
            }
        }

        public virtual DateTime firstCataloguedDateTime { get; }        

        public string resourceFullFileName { get; protected set; }
        public string resourceFullFileNameNoExt { get; protected set; }
        public string resourceFileNameExtension { get; protected set; }
        public int resourceType { get; protected set; }
        public int resourceCategory { get; protected set; }
        public int resourceTargetCategory { get; protected set; }
        public bool localFileResource { get; protected set; } = false;

        protected virtual bool vamUserHidden { get;  set; } = false;
        protected virtual bool vamUserFavourite { get;  set; } = false;   

        public virtual bool isHidden
        {
            get
            {
                return (vamUserHidden || baHidden == TriState.active);
            }
            set
            {
                if (value)
                {
                    if (baHidden != TriState.active)
                    {
                        baHidden = TriState.active;
                        SetUserDataPersist();
                    }

                }
                else
                {
                    if (baHidden != TriState.inactive)
                    {
                        baHidden = TriState.inactive;
                        SetUserDataPersist();
                    }
                }
                SetVAMUserHidden(value);
            }
        }
        public abstract bool isCSLReferenced { get; }
        public virtual bool isFavourite
        {
            get
            {
                return (vamUserFavourite || baFavourite == TriState.active);
            }
            set
            {
                if (value)
                {
                    if (baFavourite != TriState.active)
                    {
                        baFavourite = TriState.active;
                        SetUserDataPersist();
                    }

                }
                else
                {
                    if (baFavourite != TriState.inactive)
                    {
                        baFavourite = TriState.inactive;
                        SetUserDataPersist();
                    }
                }
                SetVAMUserFavourite(value);
            }
        }

        public Sprite resourceSprite = null;

        public abstract void SetUserDataPersist();

        protected abstract void SetVAMUserFavourite(bool value);
        protected abstract void SetVAMUserHidden(bool value);
        public virtual string resourceFullPathReference { get { return resourceFullFileName; }
        }
        public virtual string resourceThumbnailReference { get; }

        protected UIDynamicButton queuedButtonLoad=null;
        protected ImageLoaderThreaded.QueuedImage queuedImage = null;

        public ResourceVersionGroupEntry() { }

        protected void Init()
        {
            int fileExtensionIndex = resourceFullFileName.LastIndexOf('.');
            int fileNameIndex = 0;
            if (resourceFullFileName.Contains("\\")) fileNameIndex= resourceFullFileName.LastIndexOf('\\');
            folderName = resourceFullFileName.Substring(0, fileNameIndex);
            resourceName = resourceFullFileName.Substring(folderName.Length + 1, fileExtensionIndex - fileNameIndex - 1);
            resourceNameLowerCaps = resourceName.ToLower();
            resourceFileNameExtension = resourceFullFileName.Substring(fileExtensionIndex + 1);
            resourceFullFileNameNoExt = resourceFullFileName.Substring(0, fileExtensionIndex);

            resourceCategory = ResourceType.GetResourceCategory(resourceType);
            resourceTargetCategory = ResourceType.GetResourceTargetCategory(resourceType);
        }


        public virtual JSONClass GetCoreJSON()
        {
            JSONClass jc = new JSONClass();

            jc["resourceFullFileName"] = resourceFullFileName;
            jc["resourceType"] = EnumManifest.GetStoreVal(ResourceType.enumManifestName,resourceType);

            return (jc);
        }

        protected virtual void RestoreTags(JSONArray tagsArray)
        {
            foreach (JSONClass tagJC in tagsArray)
            {
                int tagCategory = EnumManifest.GetEnumValFromStoreVal(TagCategory.enumManifestName, tagJC["tagCategory"]);
                string tagName = tagJC["tagName"];
                if (tagCategory == TagCategory.userDefined) TagManifest.CreateTag(tagName, tagCategory, resourceType);
                TagManifest.SetRVGETagged(this, tagName, true);
            }
        }
        protected virtual JSONNode GetTagsJSON()
        {
            JSONNode ja = new JSONArray();

            foreach(var tag in TagManifest.GetRVGETags(this))
            {
                JSONClass tagJC = new JSONClass();
                tagJC["tagName"] = tag.tagName;
                tagJC["tagCategory"] = tag.tagCategoryJSEnum.storeVal;

                ja[-1] = tagJC;
            }

            return (ja);
        }

        public void Activate(ResourcePrimaryActionUI resourceActions)
        {
            if (resourceType == ResourceType.scene)
            {
                if (resourceActions.parentBrowser.isVREnabled) resourceActions.parentBrowser.SetInactive();
                if (resourceActions.mergePresetJSB.val) SuperController.singleton.LoadMerge(FileManagerSecure.NormalizePath(resourceFullPathReference));
                else SuperController.singleton.Load(FileManagerSecure.NormalizePath(resourceFullPathReference));
            }
            else if (resourceCategory == ResourceCategory.preset)
            {
                if (resourceActions.targetAtomJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.mainValFlag)
                {
                    if (AtomUtils.GetAtomNames(incudeNonPersonAtoms: false).Contains(resourceActions.targetAtomJSMultiEnum.mainVal))
                    {
                        Atom atom = SuperController.singleton.GetAtomByUid(resourceActions.targetAtomJSMultiEnum.mainVal);
                        PresetLockStore atomPresetLockStore = new PresetLockStore();
                        if (atom.type == "Person") atomPresetLockStore.StorePresetLocks(atom, true,resourceActions.suppressClothingLoadJSB.val);

                        if (resourceType == ResourceType.presetAppearance && resourceActions.suppressScaleChangeJSB.val && PC.patreonContentEnabled) PC.AppPresetSuppressScaleLoad(atom, resourceFullPathReference, resourceActions.mergePresetJSB.val);
                        else
                        {
                            JSONStorable js = null;
                            if (resourceType == ResourceType.presetAppearance) js = atom.GetStorableByID("AppearancePresets");
                            else if (PC.patreonContentEnabled) js = PC.GetPresetStorable(atom, resourceType);

                            if (js != null)
                            {
                                JSONStorableBool loadOnSelectJSB = js.GetBoolJSONParam("loadPresetOnSelect");

                                bool loadOnSelectPreState = loadOnSelectJSB.val;
                                loadOnSelectJSB.val = false;

                                bool preState1 = false;
                                bool preState2 = false;
                                bool preState3 = false;

                                if (resourceType == ResourceType.presetMorphs)
                                {
                                    JSONStorableBool includePhysical = js.GetBoolJSONParam("includePhysical");
                                    preState1 = includePhysical.val;
                                    includePhysical.val = resourceActions.morphPresetIncludePhysicalJSB.val;
                                    JSONStorableBool includeAppearance = js.GetBoolJSONParam("includeAppearance");
                                    preState2 = includeAppearance.val;
                                    includeAppearance.val = resourceActions.morphPresetIncludeAppJSB.val;
                                }
                                if (resourceType == ResourceType.presetGeneral)
                                {
                                    JSONStorableBool includePhysical = js.GetBoolJSONParam("includePhysical");
                                    preState1 = includePhysical.val;
                                    includePhysical.val = resourceActions.generalPresetIncludePhysicalJSB.val;

                                    JSONStorableBool includeAppearance = js.GetBoolJSONParam("includeAppearance");
                                    preState2 = includeAppearance.val;
                                    includeAppearance.val = resourceActions.generalPresetIncludeAppJSB.val;

                                    JSONStorableBool includePose = js.GetBoolJSONParam("includeOptional");
                                    preState3 = includePose.val;
                                    includePose.val = resourceActions.generalPresetIncludePoseJSB.val;
                                }

                                JSONStorableUrl presetPathJSON = js.GetUrlJSONParam("presetBrowsePath");
                                presetPathJSON.val = SuperController.singleton.NormalizePath(resourceFullPathReference);
                                if (resourceActions.mergePresetJSB.val) js.CallAction("MergeLoadPreset");
                                else js.CallAction("LoadPreset");

                                loadOnSelectJSB.val = loadOnSelectPreState;

                                if (resourceType == ResourceType.presetMorphs)
                                {
                                    JSONStorableBool includePhysical = js.GetBoolJSONParam("includePhysical");
                                    includePhysical.val = preState1;
                                    JSONStorableBool includeAppearance = js.GetBoolJSONParam("includeAppearance");
                                    includeAppearance.val = preState2;
                                }
                                if (resourceType == ResourceType.presetGeneral)
                                {
                                    JSONStorableBool includePhysical = js.GetBoolJSONParam("includePhysical");
                                    includePhysical.val = preState1;

                                    JSONStorableBool includeAppearance = js.GetBoolJSONParam("includeAppearance");
                                    includeAppearance.val = preState2;

                                    JSONStorableBool includePose = js.GetBoolJSONParam("includeOptional");
                                    includePose.val = preState3;
                                }
                            }                           
                        }                        
                        atomPresetLockStore.RestorePresetLocks(atom);
                    }
                }
            }
            else if (resourceCategory == ResourceCategory.plugins && PC.patreonContentEnabled)
            {
                Atom atom = null;
                if (resourceActions.targetAtomJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.mainValFlag)
                {
                    if (AtomUtils.GetAtomNames(incudeNonPersonAtoms: true).Contains(resourceActions.targetAtomJSMultiEnum.mainVal))
                    {
                        atom = SuperController.singleton.GetAtomByUid(resourceActions.targetAtomJSMultiEnum.mainVal);
                        
                    }
                }
                else if(resourceActions.targetAtomJSMultiEnum.valType == JSONStorableMultiEnumStringChooser.topEnumValFlag)
                {
                    if (resourceActions.targetAtomJSMultiEnum.valTopEnum == PluginTargetTypes.sessionPlugins) atom = Globals.mvrScript.containingAtom;
                    else atom = SuperController.singleton.GetAtomByUid("CoreControl");
                }

                PC.LoadPluginAsPreset(atom, resourceFullPathReference, resourceActions.openPluginUIJSB.val);
            }
        }


        private void QueuedImageLoadCallback(ImageLoaderThreaded.QueuedImage qi)
        {
            if (qi == queuedImage)
            {
                Texture2D tex = qi.tex;
                resourceSprite = Sprite.Create(tex, new Rect(0.0f, 0.0f, tex.width, tex.height), new Vector2(0.5f, 0.5f), 100.0f);
                LoadResourceTexture(queuedButtonLoad);
            }
            
        }

        private void LoadResourceTexture(UIDynamicButton button)
        {
            if (button != null)
            {
                Transform newImageTrans = button.button.transform.Find("ButtonSubImage");
                UnityEngine.UI.Image image = newImageTrans.GetComponent<UnityEngine.UI.Image>();
                image.sprite = resourceSprite;
            }
        }

        public void CancelButtonImageUpdate()
        {
            queuedButtonLoad = null;
        }

        public void LoadSprite(UIDynamicButton button)
        {
            string url = FileManagerSecure.NormalizePath( resourceThumbnailReference);

            Transform newImageTrans = button.button.transform.Find("ButtonSubImage");
            GameObject newImageGO = newImageTrans.gameObject;           

            UnityEngine.UI.Image image = newImageGO.GetComponent<UnityEngine.UI.Image>();

            if (string.IsNullOrEmpty(url) || !FileManagerSecure.FileExists(url))
            {
                resourceSprite = null;
                image.sprite = null;
                return;
            }
            if (resourceSprite != null)
            {
                LoadResourceTexture(button);
            }
            else 
            {
                image.sprite = null;
                queuedButtonLoad = button;
                queuedImage = ImageUtils.QueueLoadTexture(url, QueuedImageLoadCallback,true);
            }
           
        }

        public virtual bool IsTagged(Tag tag)
        {
            return TagManifest.IsRVGETagged(this, tag);
        }

        public virtual bool IsNotTagged()
        {
            return TagManifest.IsRVGENotTagged(this);
        }

        public virtual bool IsNotUserTagged()
        {
            return TagManifest.IsRVGENotUserTagged(this);
        }

        public virtual void SetTag(Tag tag, bool value)
        {
            TagManifest.SetRVGETagged(this, tag, value);
            SetUserDataPersist();
        }
    }

    public class LocalRVGE :ResourceVersionGroupEntry
    {
        public override string creatorName { get { return ""; } }
        public override string packageName { get { return ""; } }

        public override bool isCSLReferenced
        {
            get
            {
                return csListReferenced;
            }
        }
        public override DateTime firstCataloguedDateTime
        {
            get { return localFileResourceFirstScan; }
        }
        public override string resourceThumbnailReference
        {
            get
            {
                return resourceFullFileNameNoExt + ".jpg";
            }
        }
        private DateTime localFileResourceFirstScan;

        private bool persistLocalRVGE = false;
        public static bool anyPersistLocalRVGE { get; private set; } = false;

        private bool isLocalResourceSymLinkAvaliable
        {
            get
            {
                if (resourceCategory == ResourceCategory.preset && Globals.customFolderAccess) return true;
                if (resourceCategory == ResourceCategory.scene && Globals.savesFolderAccess) return true;
                return false;
            }
        }

        public LocalRVGE(string resourceFileName, int _resourceType)
        {
            resourceType = _resourceType;
            resourceFullFileName = resourceFileName;
            localFileResource = true;
            Init();

            RefreshLocalHideFavFlags();
        }

        protected override void SetVAMUserHidden(bool value)
        {
            SetVAMUserHideFav(value, ".hide");
        }
        protected override void SetVAMUserFavourite(bool value)
        {
            SetVAMUserHideFav(value, ".fav");
        }

        public bool csListReferenced = false;

        private void SetVAMUserHideFav(bool value, string fileExt)
        {
            if (isLocalResourceSymLinkAvaliable)
            {
                string path = Const.SymLinksFolder + "\\" + folderName;
                string tagFileName = path + "\\" + resourceName + "." + resourceFileNameExtension + fileExt;
                if (!FileManagerSecure.DirectoryExists(path) && value) FileManagerSecure.CreateDirectory(path);
                if (FileManagerSecure.DirectoryExists(path))
                {
                    if (value) FileManagerSecure.WriteAllBytes(tagFileName, new byte[] { 0 });
                    else FileManagerSecure.DeleteFile(tagFileName);
                }
            }
        }

        public void SetLocalFileScanTime(DateTime scanTime)
        {
            localFileResourceFirstScan = scanTime;
        }

        public override void SetUserDataPersist()
        {
            persistLocalRVGE = true;
            anyPersistLocalRVGE = true;
        }

        public void LoadLocalUserData(string fileName)
        {
            JSONClass resourceJC = SuperController.singleton.LoadJSON(fileName).AsObject;
            if (resourceJC["localFileFirstScanDateTime"] != null) localFileResourceFirstScan = DateTime.Parse(resourceJC["localFileFirstScanDateTime"], CultureInfo.InvariantCulture, DateTimeStyles.RoundtripKind);

            if (resourceJC["baFavourite"] != null) baFavourite = EnumManifest.GetEnumValFromStoreVal(TriState.enumManifestName, resourceJC["baFavourite"]);
            if (resourceJC["baHidden"] != null) baHidden = EnumManifest.GetEnumValFromStoreVal(TriState.enumManifestName, resourceJC["baHidden"]);


            if (resourceJC["Tags"] != null) RestoreTags(resourceJC["Tags"].AsArray);

            // Initial alpha release had a wrong JSON Node Name for the file format
            if (resourceJC["VARManifestStoreFormat"] != null) SetUserDataPersist();           
        }
        public void SaveLocalUserData()
        {
            if (persistLocalRVGE)
            {
                JSONClass resourceJC = GetCoreJSON();
                resourceJC["LocalUserDataStoreFormat"].AsInt = Const.LocalUserDataStoreFormat;

                resourceJC["BAMajorVersion"].AsInt = Const.BAMajorVersion;
                resourceJC["BAMinorVersion"].AsInt = Const.BAMinorVersion;
                resourceJC["BAFixVersion"].AsInt = Const.BAFixVersion;

                resourceJC["localFileFirstScanDateTime"] = localFileResourceFirstScan.ToString("O");
                resourceJC["baFavourite"] = EnumManifest.GetStoreVal(TriState.enumManifestName, baFavourite);
                resourceJC["baHidden"] = EnumManifest.GetStoreVal(TriState.enumManifestName, baHidden);

                resourceJC["Tags"]= GetTagsJSON();

                SuperController.singleton.SaveJSON(resourceJC, Const.LocalUserDataFolder + "\\" + resourceFullFileName + Const.MetaDataFileExt);
                persistLocalRVGE = false;
            }
            anyPersistLocalRVGE = false;
        }

        private void RefreshLocalHideFavFlags()
        {
            if (FileManagerSecure.FileExists(resourceFullFileName + ".hide")) vamUserHidden = true;
            else vamUserHidden = false;
            if (FileManagerSecure.FileExists(resourceFullFileName + ".fav")) vamUserFavourite = true;
            else vamUserFavourite = false;
        }
    }

    public class PackagedRVGE : ResourceVersionGroupEntry
    {
        private VARPackageVersionGroupEntry vpvge;

        private HashSet<int> varVersionsCreatorHidden = new HashSet<int>();
        private HashSet<int> varVersionsUserHidden = new HashSet<int>();
        private HashSet<int> varVersionsCreatorFav = new HashSet<int>();
        private HashSet<int> varVersionsUserFav = new HashSet<int>();
        private HashSet<int> varVersions = new HashSet<int>();
        private HashSet<int> varVersionsCSListReferenced = new HashSet<int>();
 
        //       private Dictionary<int, DateTime> varVersionFirstScanTimes = new Dictionary<int, DateTime>();

        private int highestVARVersion { get; set; }
        private bool packageCreatorHidden { get; set; } = false;
        private bool packageCreatorFav { get; set; } = false;

        public int varLicenseType { get { return vpvge.varVersionLicenses[highestVARVersion]; } }

        public override string creatorName
        {
            get {return vpvge.creatorName; }
        }
        public override string packageName
        {
            get { return vpvge.packageName; }
        }

        protected override bool vamUserHidden { get
            {
                if (varVersionsUserHidden.Count == 0) return false;
                if (varVersionsUserHidden.Max() == highestVARVersion) return true;
                return false;
            }
        }
        protected override bool vamUserFavourite { get
            {
                if (varVersionsUserFav.Count == 0) return false;
                if (varVersionsUserFav.Max() == highestVARVersion) return true;
                return false;
            }
        }

        public override bool isCSLReferenced
        {
            get
            {
                if (varVersionsCSListReferenced.Contains(highestVARVersion)) return true;
                return false;
            }
        }

        public override DateTime firstCataloguedDateTime
        {
            get {
               return vpvge.firstCataloguedDateTime; }
        }

        public override bool resourcePresent
        {
            get
            {
                if (vpvge.highestVARVersion == highestVARVersion) return true;
                return false;
            }
        }

        public override bool isHidden
        {
            get
            {
                return (vamUserHidden || packageCreatorHidden );
            }
        }

        public override bool isFavourite
        {
            get
            {
                return (vamUserFavourite || packageCreatorFav );
            }
        }

        public override void SetUserDataPersist()
        {
            vpvge.SetPersistVARUserData();
        }

        protected override void SetVAMUserFavourite(bool value)
        {
            bool success=SetVAMUserHideFav(value, ".fav");
            if (success) {
                if (value) varVersionsUserFav.Add(highestVARVersion);
                else varVersionsUserFav.Remove(highestVARVersion);
            }
        }
        protected override void SetVAMUserHidden(bool value)
        {
            bool success = SetVAMUserHideFav(value, ".hide");
            if (success)
            {
                if (value) varVersionsUserHidden.Add(highestVARVersion);
                else varVersionsUserHidden.Remove(highestVARVersion);
            }
        }

        private bool SetVAMUserHideFav(bool value, string fileExt)
        {
            if (Globals.addonPackageFilePrefsFolderAccess)
            {
                string path = Const.SymLinksFolder + "\\AddonPackagesFilePrefs\\" + creatorName + "." + packageName + "." + highestVARVersion.ToString() + "\\" + folderName;
                string tagFileName = path + "\\" + resourceName + "." + resourceFileNameExtension + fileExt;
                if (!FileManagerSecure.DirectoryExists(path) && value) FileManagerSecure.CreateDirectory(path);
                if (FileManagerSecure.DirectoryExists(path))
                {
                    if (value) FileManagerSecure.WriteAllBytes(tagFileName, new byte[] { 0 });
                    else FileManagerSecure.DeleteFile(tagFileName);
                }
                return true;
            }
            return false;
        }

        public override string resourceFullPathReference
        {
            get
            {
                return creatorName + "." + packageName + "." + highestVARVersion.ToString() + ":\\" + resourceFullFileName;
            }

        }
        public override string resourceThumbnailReference
        {
            get
            {
                return creatorName + "." + packageName + "." + highestVARVersion.ToString() + ":\\" + resourceFullFileNameNoExt + ".jpg";
            }
        }

        public PackagedRVGE(VARPackageManifestEntry vpme, string resourceFileName, int _resourceType) 
        {
            vpvge = vpme.vpvge;
            resourceType = _resourceType;
            resourceFullFileName = resourceFileName;


            highestVARVersion = vpme.varVersion;

            varVersions.Add(vpme.varVersion);

            Init();
        }

        public PackagedRVGE(JSONClass jc, VARPackageVersionGroupEntry _vpvge)
        {
            vpvge = _vpvge;

            if (jc["resourceFullFileName"] != null) resourceFullFileName = jc["resourceFullFileName"];
            if (jc["resourceType"] != null) resourceType = EnumManifest.GetEnumValFromStoreVal(ResourceType.enumManifestName, jc["resourceType"]);

            JSONArray varVersionsJA = jc["varVersions"].AsArray;
            foreach (JSONNode jn in varVersionsJA) AddVARVersion(jn.AsInt);
            JSONArray varVersionsCreatorHidJA = jc["varCreatorHidVersions"].AsArray;
            foreach (JSONNode jn in varVersionsCreatorHidJA) varVersionsCreatorHidden.Add(jn.AsInt);
            JSONArray varVersionsCreatorFavJA = jc["varCreatorFavVersions"].AsArray;
            foreach (JSONNode jn in varVersionsCreatorFavJA) varVersionsCreatorFav.Add(jn.AsInt);

            if (jc["varVersionsCSLReferenced"] != null)
            {
                JSONArray varVersionsCSLReferencedJA = jc["varVersionsCSLReferenced"].AsArray;
                foreach (JSONNode jn in varVersionsCSLReferencedJA) varVersionsCSListReferenced.Add(jn.AsInt);
            }           

            Init();
        }

        public void RestoreVARUserData(JSONClass resourceJC)
        {

            if (resourceJC["baFavourite"] != null) baFavourite = EnumManifest.GetEnumValFromStoreVal(TriState.enumManifestName, resourceJC["baFavourite"]);
            if (resourceJC["baHidden"] != null) baHidden = EnumManifest.GetEnumValFromStoreVal(TriState.enumManifestName, resourceJC["baHidden"]);

            if (resourceJC["Tags"] != null) RestoreTags(resourceJC["Tags"].AsArray);
        }

        public JSONClass GetVARUserDataJSON()
        {
            JSONClass resourceJC = new JSONClass();
            resourceJC["resourceFullFileName"] = resourceFullFileName;         

            resourceJC["baFavourite"] = EnumManifest.GetStoreVal(TriState.enumManifestName, baFavourite);
            resourceJC["baHidden"] = EnumManifest.GetStoreVal(TriState.enumManifestName, baHidden);

            resourceJC["Tags"] = GetTagsJSON();

            return resourceJC;
        }

        public override JSONClass GetCoreJSON()
        {
            JSONClass jc = base.GetCoreJSON();

            JSONArray varVersionsJA = new JSONArray();
            jc["varVersions"] = varVersionsJA;
            foreach (int varVersion in varVersions) varVersionsJA[-1].AsInt = varVersion;

            JSONArray varVersionsCreatorHidJA = new JSONArray();
            jc["varCreatorHidVersions"] = varVersionsCreatorHidJA;
            foreach (int varVersion in varVersionsCreatorHidden) varVersionsCreatorHidJA[-1].AsInt = varVersion;

            JSONArray varVersionsCreatorFavJA = new JSONArray();
            jc["varCreatorFavVersions"] = varVersionsCreatorFavJA;
            foreach (int varVersion in varVersionsCreatorFav) varVersionsCreatorFavJA[-1].AsInt = varVersion;

            JSONArray varVersionsCSLReferencedJA = new JSONArray();
            jc["varVersionsCSLReferenced"] = varVersionsCSLReferencedJA;
            foreach (int varVersion in varVersionsCSListReferenced) varVersionsCSLReferencedJA[-1].AsInt = varVersion;

            return (jc);
        }

        public void AddVARVersion(int varVersion)
        {
            varVersions.Add(varVersion);

            if (varVersion > highestVARVersion)
            {
                highestVARVersion = varVersion;
                if (varVersionsCreatorFav.Contains(varVersion)) packageCreatorFav = true;
                else packageCreatorFav = false;
                if (varVersionsCreatorHidden.Contains(varVersion)) packageCreatorHidden = true;
                else packageCreatorHidden = false;
            }
        }

        public void NewVARVersions(HashSet<int> newVARVersions)
        {            
            foreach (int varVersion in newVARVersions.Union(varVersions))
            {
                if (baHidden == TriState.active) SetVAMUserHidden(true);
                if (baFavourite == TriState.active) SetVAMUserFavourite(true);
            }
        }

        public void SetHideCreatorVARVersion(int varVersion)
        {
            varVersionsCreatorHidden.Add(varVersion);
            if (varVersion == highestVARVersion) packageCreatorHidden = true;
        }

        public void SetFavCreatorVARVersion(int varVersion)
        {
            varVersionsCreatorFav.Add(varVersion);
            if (varVersion == highestVARVersion) packageCreatorFav = true;
        }

        public void SetHideUserVARVersion(int varVersion, bool hide)
        {
            if (hide) varVersionsUserHidden.Add(varVersion);
            else varVersionsUserHidden.Remove(varVersion);
            if (varVersion == highestVARVersion)
            {
                if (hide)
                {
                    if (baHidden != TriState.active)
                    {
                        baHidden = TriState.active;
                        vpvge.SetPersistVARUserData();
                    }
                    
                }
                else
                {
                    if (baHidden != TriState.inactive)
                    {
                        baHidden = TriState.inactive;
                        vpvge.SetPersistVARUserData();
                    }
                }
            }
        }
        
        public void SetFavUserVARVersion(int varVersion, bool fav)
        {
            if (fav) varVersionsUserFav.Add(varVersion);
            else varVersionsUserFav.Remove(varVersion);
            if (varVersion == highestVARVersion)
            {
                if (fav)
                {
                    if (baFavourite != TriState.active)
                    {
                        baFavourite = TriState.active;
                        vpvge.SetPersistVARUserData();
                    }
                    
                }
                else
                {
                    if (baFavourite != TriState.inactive)
                    {
                        baFavourite = TriState.inactive;
                        vpvge.SetPersistVARUserData();
                    }
                }
            }
        }

        public void ClearAllUserFilePrefs()
        {
            varVersionsUserHidden.Clear();
            varVersionsUserFav.Clear();
        }

        public void SetCSLFileReferenced(int varVersion)
        {
            varVersionsCSListReferenced.Add(varVersion);
        }
    }

    public class CreatorManifest
    {
        private static Dictionary<string, CreatorManifestEntry> creatorManifestDict = new Dictionary<string, CreatorManifestEntry>();

        public static CreatorManifestEntry RegisterCreator (string creatorName)
        {
            if (creatorManifestDict.ContainsKey(creatorName)) return creatorManifestDict[creatorName];
            CreatorManifestEntry cme = new CreatorManifestEntry(creatorName);
            creatorManifestDict.Add(creatorName,cme);
            return cme;
        }

        public static void RegisterResource (ResourceVersionGroupEntry rvge)
        {
            creatorManifestDict[rvge.creatorName].RegisterResource(rvge);
        }


        public static HashSet<string> GetAllCreatorNames()
        {
            return new HashSet<string>(creatorManifestDict.Keys);
        }

        public static List<string> GetFilteredCreatorNames(BrowserFiltersUI filters)
        {
            var filteredCreators = creatorManifestDict.Values.Where(o => o.GetResourceCount(filters) > 0).Select(o => o.creatorName).OrderBy(s => s).ToList();
            return filteredCreators;
        }

        public static List<string> GetFilteredCreatorNamesWithResourceCount(BrowserFiltersUI filters)
        {
            var filteredCreators = creatorManifestDict.Values.Where(o => o.GetResourceCount(filters) > 0).Select(o => o.creatorName + " (" + o.GetResourceCount(filters).ToString() + ")").OrderBy(s => s).ToList();

            return filteredCreators;
        }
    }

    public class CreatorManifestEntry
    {
        public string creatorName { get; private set; }

        public HashSet<int> resourceTypes { get; private set; }

        public Dictionary<int, HashSet<ResourceVersionGroupEntry>> resourcesByTypeDict = new Dictionary<int, HashSet<ResourceVersionGroupEntry>>();
        public CreatorManifestEntry(string _creatorName)
        {
            creatorName = _creatorName;
            resourceTypes = new HashSet<int>();
        }

        public int GetResourceCount(BrowserFiltersUI filter)
        {
            
            return ResourceManifest.GetResourceGroupsByCreator(filter.resourceTypeJSEnum.val, filter.showHiddenJSB.val, filter.onlyFavouritesJSB.val, filter.tagsFilterJSMultiEnum, creatorName).Count;
        }

        public void RegisterResource (ResourceVersionGroupEntry rvge)
        {
            resourceTypes.Add(rvge.resourceType);

            if (!resourcesByTypeDict.ContainsKey(rvge.resourceType)) resourcesByTypeDict.Add(rvge.resourceType, new HashSet<ResourceVersionGroupEntry>());
            resourcesByTypeDict[rvge.resourceType].Add(rvge);
        }
    }
}
