using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using MVR.FileManagementSecure;
using UnityEngine.UI;

namespace JayJayWon
{
    public class ResourceActionsUI
    {
        private GameObject parentBoxGO;
        private GameObject actionsUIBoxGO;
        private GameObject tabsUIBoxGO;
        private BrowserUI parentBrowser;

        private Dictionary<int, ResourceActionUIBase> resourceActionUITabs = new Dictionary<int, ResourceActionUIBase>();

        private List<UIDynamicButton> resourceActionTabButtons = new List<UIDynamicButton>();

        public int currentActionUITabType = ActionUITabType.primaryActionTab;

        const float resourceActionTabsHeight = 15f;
        const float resourceActionTabsWidth = 50f;

        public ResourceActionUIBase currentActionUITab
        {
            get
            {
                List<int> actionChoices = GetActionUITabChoices();
                if (!actionChoices.Contains(currentActionUITabType))
                {
                    if (resourceActionUITabs.ContainsKey(currentActionUITabType)) resourceActionUITabs[currentActionUITabType].ClearUI();
                    currentActionUITabType = actionChoices[0];
                    RefreshActionTabs();
                }
                if (resourceActionUITabs.ContainsKey(currentActionUITabType)) return resourceActionUITabs[currentActionUITabType];
                return CreateActionTab(currentActionUITabType);
            }
        }
        public ResourcePrimaryActionUI primaryActionUITab
        {
            get
            {
                List<int> actionChoices = GetActionUITabChoices();

                if (resourceActionUITabs.ContainsKey(ActionUITabType.primaryActionTab)) return (ResourcePrimaryActionUI)resourceActionUITabs[ActionUITabType.primaryActionTab];
                return (ResourcePrimaryActionUI)CreateActionTab(ActionUITabType.primaryActionTab);
            }
        }
        private void RefreshActionTabs()
        {
            var actionTabTypes = GetActionUITabChoices();
            int currentTabButtonCount = resourceActionTabButtons.Count;
            int requiredTabButtonCount = actionTabTypes.Count;

            if (currentTabButtonCount < requiredTabButtonCount)
            {
                for (int i = currentTabButtonCount; i < requiredTabButtonCount; i++) resourceActionTabButtons.Add(CreateResourceActionTabButton(i));
            }
            else if (currentTabButtonCount > requiredTabButtonCount)
            {
                for (int i = requiredTabButtonCount; i < currentTabButtonCount; i++)
                {
                    GameObject.Destroy(resourceActionTabButtons[i].gameObject);
                    resourceActionTabButtons.RemoveRange(requiredTabButtonCount, (currentTabButtonCount - requiredTabButtonCount));
                }
            }
            int index = 0;
            foreach (var tab in resourceActionTabButtons)
            {
                tab.label = ActionUITabType.GetTabLabel(actionTabTypes[index]);
                int actionTabType = actionTabTypes[index];
                if (actionTabType == currentActionUITabType)
                {
                    tab.buttonColor = BrowserUI.darkBlueGreyTransparent;
                    tab.textColor = Color.white;
                }
                else
                {
                    tab.buttonColor = BrowserUI.lightGrey;
                    tab.textColor = Color.black;
                }
                var buttonUID = resourceActionTabButtons[index];
                buttonUID.button.onClick.RemoveAllListeners();
                buttonUID.button.onClick.AddListener(delegate {
                    TabSelected(actionTabType);
                });
                AssignTabHelpText(buttonUID, actionTabType);
                index++;
            }
        }

        private void AssignTabHelpText(UIDynamicButton buttonUID, int actionTabType)
        {
            string helpText = "";

            if (actionTabType == ActionUITabType.primaryActionTab) helpText = "Perform primary actions on the selected resource. Where multiple resource are selected, the primary action will be performed on the last selected resource.";
            if (actionTabType == ActionUITabType.tagsTab) helpText = "Assign tags to the selected resources.";
            parentBrowser.ApplyHelp(buttonUID, helpText);
        }

        private void TabSelected(int tabType)
        {
            if (currentActionUITabType != tabType)
            {
                currentActionUITab.ClearUI();
                currentActionUITabType = tabType;
                RefreshUI();
            }
        }
        private UIDynamicButton CreateResourceActionTabButton(int i)
        {
            UIDynamicButton newTabButton = UI.CreateUIButton(tabsUIBoxGO, resourceActionTabsWidth, resourceActionTabsHeight, i.ToString(), 20, i * resourceActionTabsWidth, anchorPreset: AnchorPresets.bottomLeft);
            newTabButton.buttonImage.sprite = null;
            return newTabButton;
        }

        public void RefreshUI()
        {
            RefreshActionTabs();
            currentActionUITab.RefreshUI();
        }

        public ResourceActionsUI(BrowserUI _parentBrowser, GameObject _parentBoxGO)
        {
            parentBrowser = _parentBrowser;
            parentBoxGO = _parentBoxGO;

            tabsUIBoxGO = UI.AddChildGOImage(parentGO: _parentBoxGO,
                color: Color.grey,
                anchorPreset: AnchorPresets.topMiddle,
                horizontalSize: 640f,
                verticalSize: resourceActionTabsHeight,
                anchoredPositionOffset: new Vector2(0f, 0f));

            actionsUIBoxGO = UI.AddChildGOImage(parentGO: _parentBoxGO,
                color: BrowserUI.darkBlueGreyTransparent,
                anchorPreset: AnchorPresets.bottomMiddle,
                horizontalSize: 640f,
                verticalSize: BrowserUI.resourceActionBoxHeight - resourceActionTabsHeight,
                anchoredPositionOffset: new Vector2(0f, 0f));

        }

        public void AtomNameUpdated(string oldName, string newName)
        {
            foreach (var kvp in resourceActionUITabs) kvp.Value.AtomNameUpdated(oldName, newName);
        }

        public void OnAtomAdded(Atom atom)
        {
            if (BrowserAssist.browserUI.isUIActive) primaryActionUITab.RefreshTargetAtomValue();

        }

        public void OnAtomRemoved(Atom atom)
        {
            if (BrowserAssist.browserUI.isUIActive) primaryActionUITab.RefreshTargetAtomValue();

        }

        private ResourceActionUIBase CreateActionTab(int actionTabType)
        {
            ResourceActionUIBase newActionTab = null;
            if (actionTabType == ActionUITabType.primaryActionTab) newActionTab = new ResourcePrimaryActionUI(parentBrowser, actionsUIBoxGO);
            if (actionTabType == ActionUITabType.tagsTab) newActionTab = new ResourceTagsActionUI(parentBrowser, actionsUIBoxGO);

            if (newActionTab != null)
            {
                if (resourceActionUITabs.ContainsKey(actionTabType)) resourceActionUITabs[actionTabType] = newActionTab;
                else resourceActionUITabs.Add(actionTabType, newActionTab);
            }
            return (newActionTab);
        }

        private List<int> GetActionUITabChoices()
        {
            List<int> choices = new List<int>();

            choices.Add(ActionUITabType.primaryActionTab);
            choices.Add(ActionUITabType.tagsTab);
            return choices;
        }

    }


    public abstract class ResourceActionUIBase
    {
        protected GameObject parentBoxGO;
        public BrowserUI parentBrowser { get; protected set; }

        protected const float edgeWidth = 2f;

        protected HashSet<UIDynamic> uiElements = new HashSet<UIDynamic>();

        public ResourceActionUIBase(BrowserUI _parentBrowser, GameObject _parentBoxGO)
        {
            parentBrowser = _parentBrowser;
            parentBoxGO = _parentBoxGO;

        }

        public virtual void AtomNameUpdated(string oldName, string newName)
        {
        }

        public virtual void RefreshUI()
        {
            ClearUI();
        }

        public virtual void ClearUI()
        {
            foreach (var uiElement in uiElements) if (uiElement!=null) GameObject.Destroy(uiElement.gameObject);
            uiElements.Clear();
        }

        public UIDynamicButton CreateUIButton(GameObject parentGO, float width, float height, string label = "", int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none)
        {
            UIDynamicButton button = UI.CreateUIButton(parentGO, width, height, label, fontSize, anchorXOffset, anchorYOffset, anchorPreset);
            uiElements.Add(button);
            return button;
        }

        public UIDynamicPopup CreateScrollablePopup(GameObject parentGO, JSONStorableStringChooser jssc, float width, float height, string label = "", int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, float popupHeight = 300f, bool popupAbove = false, bool disableSlider=false, int anchorPreset = AnchorPresets.none)
        {
            UIDynamicPopup popup = UI.CreateScrollablePopup(parentGO, jssc, width, height, label, fontSize, anchorXOffset, anchorYOffset, popupHeight,popupAbove, disableSlider, anchorPreset);
            uiElements.Add(popup);
            return popup;
        }

        public UIDynamicPopup CreatePopup(GameObject parentGO, JSONStorableStringChooser jssc, float width, float height, string label = "", int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool popAbove = false)
        {
            UIDynamicPopup popup = UI.CreatePopup(parentGO, jssc, width, height, label, fontSize, anchorXOffset, anchorYOffset, anchorPreset, popAbove);
            uiElements.Add(popup);
            return popup;
        }

        public UIDynamicTextField CreateTextField(GameObject parentGO, JSONStorableString jss, float width, float height, int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool input = false, InputField.LineType inputFieldType = InputField.LineType.MultiLineNewline, string placeHolderText = "", float hScrollBarHeight = 0f, float vScrollBarWidth = 10f, Action<string> textEntryChangeCallback = null)
        {
            UIDynamicTextField textField = UI.CreateTextBox(parentGO, jss, width, height, fontSize, anchorXOffset, anchorYOffset, anchorPreset, input, inputFieldType, placeHolderText, hScrollBarHeight, vScrollBarWidth, textEntryChangeCallback);
            uiElements.Add(textField);
            return textField;
        }

        public UIDynamicToggle CreateToggle(GameObject parentGO, JSONStorableBool jsb, float width, float height, string label = "", int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none)
        {
            UIDynamicToggle toggle = UI.CreateToggle(parentGO, jsb, width, height, label, fontSize, anchorXOffset, anchorYOffset, anchorPreset);
            uiElements.Add(toggle);
            return toggle;
        }
        public TriStateToggle CreateTriStateToggle(GameObject parentGO, JSONStorableBool jsb, float width, float height, string label = "", int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none)
        {
            UIDynamicToggle toggle = UI.CreateToggle(parentGO, jsb, width, height, label, fontSize, anchorXOffset, anchorYOffset, anchorPreset);
            uiElements.Add(toggle);
            return new TriStateToggle(jsb, toggle);
        }


    }

    public class ResourcePrimaryActionUI : ResourceActionUIBase
    {
        public JSONStorableMultiEnumStringChooser targetAtomJSMultiEnum;
        public JSONStorableBool mergePresetJSB;
        public JSONStorableBool openPluginUIJSB;
        public JSONStorableBool suppressClothingLoadJSB;
        public JSONStorableBool suppressScaleChangeJSB;

        public JSONStorableBool morphPresetIncludePhysicalJSB;
        public JSONStorableBool morphPresetIncludeAppJSB;

        public JSONStorableBool generalPresetIncludePhysicalJSB;
        public JSONStorableBool generalPresetIncludePoseJSB;
        public JSONStorableBool generalPresetIncludeAppJSB;

        private int resourceType { get { return parentBrowser.browserFiltersUI.resourceTypeJSEnum.val; } }     
        private int resourceCategory { get { return ResourceType.GetResourceCategory(resourceType); } }

        UIDynamicPopup targetAtomPopup;
        UIDynamicButton mainActionButton;

        public ResourcePrimaryActionUI(BrowserUI _parentBrowser, GameObject _parentBoxGO) : base(_parentBrowser, _parentBoxGO)
        {
            targetAtomJSMultiEnum = new JSONStorableMultiEnumStringChooser("targetAtom", null, null, null, null, NoneAvailable.enumManifestName, null);
            targetAtomJSMultiEnum.label = "Target Atom";

            mergePresetJSB = new JSONStorableBool("mergePreset", false);
            openPluginUIJSB = new JSONStorableBool("openPluginUI", false);
            suppressClothingLoadJSB = new JSONStorableBool("suppressClothingLoad", false);
            suppressScaleChangeJSB = new JSONStorableBool("suppressClothingLoad", false);

            morphPresetIncludePhysicalJSB = new JSONStorableBool("morphPresetIncludePhysical", false);
            morphPresetIncludeAppJSB = new JSONStorableBool("morphPresetIncludeApp", true);

            generalPresetIncludePhysicalJSB = new JSONStorableBool("generalPresetIncludePhysical", true);
            generalPresetIncludePoseJSB = new JSONStorableBool("generalPresetIncludePose", true);
            generalPresetIncludeAppJSB = new JSONStorableBool("generalPresetIncludeApp", true);
        }

        public override void AtomNameUpdated(string oldName, string newName)
        {
            if (targetAtomJSMultiEnum.valType != JSONStorableMultiEnumStringChooser.topEnumValFlag)
            {
                if (targetAtomJSMultiEnum.mainVal == oldName)
                {
                    ResetTargetAtomChoices(GetTargetAtomNames());
                    targetAtomJSMultiEnum.mainVal = newName;
                }
            }
        }

        public override void RefreshUI()
        {
            base.RefreshUI();           

            if (resourceType==ResourceType.scene || resourceCategory == ResourceCategory.preset)
            {
                var mergeToggle = CreateToggle(parentBoxGO, mergePresetJSB, 80f, 20f, "Merge Load", 20, edgeWidth, -edgeWidth,AnchorPresets.topLeft);
            }

            if (resourceType == ResourceType.plugins)
            {
                var openPluginUIToggle = CreateToggle(parentBoxGO, openPluginUIJSB, 100f, 20f, "Open Plugin UI", 20, edgeWidth, -edgeWidth, AnchorPresets.topLeft);
            }
            if (resourceCategory == ResourceCategory.preset || resourceType == ResourceType.plugins)
            {
                targetAtomPopup = CreateScrollablePopup(parentBoxGO, targetAtomJSMultiEnum, 160f, 30f, "Target Atom", 20,edgeWidth, -edgeWidth-20f,300f, true, true, AnchorPresets.topLeft);
                targetAtomPopup.popup.onOpenPopupHandlers += RefreshTargetAtomPopup;
                parentBrowser.ApplyHelp(targetAtomPopup, "Select a target atom from the current scene that the primary action for the selected resource will be performed on.");
                RefreshTargetAtomValue();
            }

            
            if (resourceCategory == ResourceCategory.preset)
            {
                string resourceTypeName = parentBrowser.browserFiltersUI. resourceTypeJSEnum.displayVal.Substring(7);
                mainActionButton = CreateUIButton(parentBoxGO, 100f, 30f, "Load " + resourceTypeName + " Preset", 20, 165f+ edgeWidth, -edgeWidth-20f, AnchorPresets.topLeft);
            }
            else if (resourceType == ResourceType.scene)
            {
                mainActionButton = CreateUIButton(parentBoxGO, 100f, 30f, "Load Scene", 20, edgeWidth, -edgeWidth-20f, AnchorPresets.topLeft);
            }
            else if (resourceType == ResourceType.plugins)
            {
                mainActionButton = CreateUIButton(parentBoxGO, 100f, 30f, "Load Plugin", 20, 165f + edgeWidth, -edgeWidth-20f, AnchorPresets.topLeft);
            }

            if (mainActionButton != null)
            {
                mainActionButton.buttonColor = new Color(0, 0.6f, 0);
                mainActionButton.textColor = Color.white;
                mainActionButton.button.onClick.AddListener(delegate
                {
                    ResourceVersionGroupEntry rvge = parentBrowser.resourceSelectionsUI.GetFirstSelectedRVGE(resourceType);
                    if (rvge != null) rvge.Activate(this);

                });
                PrimaryActionPointerEnterExitHelp primaryActionContextHelp = mainActionButton.button.gameObject.AddComponent<PrimaryActionPointerEnterExitHelp>();
                primaryActionContextHelp.Init(parentBrowser.helpTextJSS, parentBrowser.resourceSelectionsUI.GetResourceTypeSelections(resourceType), this);
            }
            if (resourceType == ResourceType.presetAppearance)
            {
                var suppressClothingLoadToggle = CreateToggle(parentBoxGO, suppressClothingLoadJSB, 140f, 20f, "Suppress Clothing Load", 20, edgeWidth, -edgeWidth - 50, AnchorPresets.topLeft);

                if (PC.patreonContentEnabled)
                {
                    var suppressScaleChangeToggle = CreateToggle(parentBoxGO, suppressScaleChangeJSB, 140f, 20f, "Suppress Scale Change", 20, edgeWidth + 140f, -edgeWidth - 50, AnchorPresets.topLeft);
                }
            }
            else if(resourceType == ResourceType.presetMorphs)
            {
                var morphPresetIncludeAppToggle = CreateToggle(parentBoxGO, morphPresetIncludeAppJSB, 120f, 20f, "Include Appearance", 20, edgeWidth, -edgeWidth - 50, AnchorPresets.topLeft);
                var morphPresetIncludePhysical = CreateToggle(parentBoxGO, morphPresetIncludePhysicalJSB, 140f, 20f, "Include Physical (Pose)", 20, edgeWidth + 120f, -edgeWidth - 50, AnchorPresets.topLeft);
            }
            else if (resourceType == ResourceType.presetGeneral)
            {
                var generalPresetIncludePhysicalToggle = CreateToggle(parentBoxGO, generalPresetIncludePhysicalJSB, 120f, 20f, "Include Physical", 20, edgeWidth, -edgeWidth - 50, AnchorPresets.topLeft);
                var generalPresetIncludePoseToggle = CreateToggle(parentBoxGO, generalPresetIncludePoseJSB, 100f, 20f, "Include Pose", 20, edgeWidth+120f, -edgeWidth - 50, AnchorPresets.topLeft);
                var generalPresetIncludeAppToggle = CreateToggle(parentBoxGO, generalPresetIncludeAppJSB, 125f, 20f, "Include Appearance", 20, edgeWidth+220f, -edgeWidth - 50, AnchorPresets.topLeft);
            }
        }

        public void RefreshTargetAtomValue()
        {
            List<string> targetAtomNames = GetTargetAtomNames();
            ResetTargetAtomChoices(targetAtomNames);

            if (targetAtomJSMultiEnum.valType != JSONStorableMultiEnumStringChooser.topEnumValFlag || resourceType != ResourceType.plugins)
            {
                if (targetAtomNames.Count == 0)
                {
                    if (resourceType == ResourceType.plugins) targetAtomJSMultiEnum.valTopEnum = PluginTargetTypes.sessionPlugins;
                    else targetAtomJSMultiEnum.valTopEnum = NoneAvailable.noneAvailable;
                }
                else
                {
                    if (!targetAtomNames.Contains(targetAtomJSMultiEnum.mainVal))
                    {
                        targetAtomJSMultiEnum.mainVal = targetAtomNames[0];
                    }
                }
            }
            
        }

        private List<string> GetTargetAtomNames()
        {
            List<string> targetAtomNames = new List<string>();
            int resoureTargetCategory = ResourceType.GetResourceTargetCategory(resourceType);

            if (resoureTargetCategory != ResourceTargetCategory.nonAtomTargetable)
            {
                List<Atom> atoms;

                if (resoureTargetCategory == ResourceTargetCategory.personAtomTargetable) atoms = AtomUtils.GetAtoms(incudeNonPersonAtoms: false);
                else if (resoureTargetCategory == ResourceTargetCategory.otherAtomTargetable) atoms = AtomUtils.GetAtoms(includePersonAtoms: false);
                else atoms = AtomUtils.GetAtoms();

                foreach (Atom atom in atoms) targetAtomNames.Add(atom.name);
            }
            targetAtomNames.Sort();
            return targetAtomNames;
        }

        private void RefreshTargetAtomPopup()
        {
            ResetTargetAtomChoices(GetTargetAtomNames());
        }

        private void ResetTargetAtomChoices(List<string> atomNames)
        {
            if (atomNames.Count == 0)
            {
                targetAtomJSMultiEnum.SetMainChoices(null);
                if (resourceType== ResourceType.plugins) targetAtomJSMultiEnum.SetTopEnum(PluginTargetTypes.enumManifestName);
                else targetAtomJSMultiEnum.SetTopEnum(NoneAvailable.enumManifestName);
            }
            else
            {
                if (resourceType == ResourceType.plugins)  targetAtomJSMultiEnum.SetTopEnum(PluginTargetTypes.enumManifestName);
                else targetAtomJSMultiEnum.SetTopEnum(null);
                targetAtomJSMultiEnum.SetMainChoices(atomNames);
            }
            if (targetAtomJSMultiEnum.popup != null)
            {
                if (atomNames.Count > 10) targetAtomJSMultiEnum.popup.popupPanelHeight = 250f;
                else targetAtomJSMultiEnum.popup.popupPanelHeight = (atomNames.Count + 1) * 25f;
            }
            
        }
    }

    public class ResourceTagsActionUI : ResourceActionUIBase
    {
        private JSONStorableBool hiddenJSB;
        private JSONStorableBool favouriteJSB;

        private TriStateToggle hideToggle;
        private TriStateToggle favToggle;

        private string tagNameFilter = "";

        private JSONStorableString tagFilterJSS;
        private JSONStorableEnumStringChooser tagCategoryFilterJSEnum;

        private ResourceSelections resourceSelectionsUI { get { return parentBrowser.resourceSelectionsUI; } }

        private BrowserFiltersUI browserFiltersUI { get { return parentBrowser.browserFiltersUI; } }

        private int filtersResourceType { get { return browserFiltersUI.resourceTypeJSEnum.val; } }

        private int filtersSelectionMode { get { return browserFiltersUI.selectModeJSEnum.val; } }

        private GameObject tagSelectionsScrollableBoxGO;
        private GameObject tagSelectionScrollableViewportContentGO;
        private List<GameObject> tagSelectionRows = new List<GameObject>();
        private UIDynamicButton createNewTagButton;
        private UIDynamicButton clearTagFilterButton;

        private Dictionary<JSONStorableBool, TriStateToggle> tagJSBtoTriStateToggle = new Dictionary<JSONStorableBool, TriStateToggle>();
        private Dictionary<JSONStorableBool, Tag> tagJSBtoTag = new Dictionary<JSONStorableBool, Tag>();

        public ResourceTagsActionUI(BrowserUI _parentBrowser, GameObject _parentBoxGO) : base(_parentBrowser, _parentBoxGO)
        {
            hiddenJSB = new JSONStorableBool("hidden", false, HiddenCallback);
            favouriteJSB = new JSONStorableBool("favourite", false, FavouriteCallback);
            tagFilterJSS = new JSONStorableString("tagFilter", "");
            tagCategoryFilterJSEnum = new JSONStorableEnumStringChooser("tagCategoryFilter", TagCategory.enumManifestName, TagCategory.all, "Tag Category");
        }

        public override void ClearUI()
        {
            base.ClearUI();
            GameObject.Destroy(tagSelectionsScrollableBoxGO);
        }

        public override void RefreshUI()
        {
            base.RefreshUI();

            ClearTagSelectionRows();

            tagJSBtoTriStateToggle.Clear();
            tagJSBtoTag.Clear();
                
            if (resourceSelectionsUI.SelectionCount(filtersResourceType)>0)
            {
                hideToggle = CreateTriStateToggle(parentBoxGO, hiddenJSB, 75, 20, "Hide", 20, -2, -3, anchorPreset: AnchorPresets.topRight);
                HiddenPointerEnterExitHelp hiddenContextHelp = hideToggle.toggle.gameObject.AddComponent<HiddenPointerEnterExitHelp>();
                hiddenContextHelp.Init(parentBrowser.helpTextJSS, resourceSelectionsUI.GetResourceTypeSelections(browserFiltersUI.resourceTypeJSEnum.val), browserFiltersUI);

                //               parentBrowser.ApplyHelp(hideToggle.toggleUID, "Toggle the VAM Hidden state for the selected resources.");
                favToggle = CreateTriStateToggle(parentBoxGO, favouriteJSB, 75, 20, "Favorite", 20, -2, -25, anchorPreset: AnchorPresets.topRight);
                FavPointerEnterExitHelp favContextHelp = favToggle.toggle.gameObject.AddComponent<FavPointerEnterExitHelp>();
                favContextHelp.Init(parentBrowser.helpTextJSS, resourceSelectionsUI.GetResourceTypeSelections(browserFiltersUI.resourceTypeJSEnum.val), browserFiltersUI);
                //               parentBrowser.ApplyHelp(favToggle.toggleUID, "Toggle the VAM Favorite state for the selected resources.");

                if (!Globals.addonPackageFilePrefsFolderAccess || !Globals.customFolderAccess || !Globals.savesFolderAccess)
                {
                    favToggle.toggle.interactable = false;
                    hideToggle.toggle.interactable = false;

                    parentBrowser.ApplyHelp(hideToggle.toggleUID, "Enable BA Sym Links using .bat file provided in PluginData folder to allow toggling the VAM Hidden state from within Browser Assist.");
                    parentBrowser.ApplyHelp(favToggle.toggleUID, "Enable BA Sym Links using .bat file provided in PluginData folder to allow toggling the VAM Favorite state from within Browser Assist.");
                }

                hideToggle.RefreshToggle(resourceSelectionsUI.IsSelectionHidden(filtersResourceType, filtersSelectionMode));

                favToggle.RefreshToggle(resourceSelectionsUI.IsSelectionFavourite(filtersResourceType, filtersSelectionMode));

                var textFieldUID = CreateTextField(parentBoxGO, tagFilterJSS, 150f, 20f, 20, 2, -2, AnchorPresets.topLeft, true, InputField.LineType.SingleLine, "Filter or create new tag...", 0, 0,TagFilterNameCallback);
                parentBrowser.ApplyHelp(textFieldUID, "Enter text to filter the list of Tags that are available for this resource type. Create new Tags, where an existing Tag does not match the entered search text.");

                clearTagFilterButton = CreateUIButton(parentBoxGO, 16f, 16f, "X", 20, 134f, -4, AnchorPresets.topLeft);
                clearTagFilterButton.button.onClick.AddListener(delegate { tagFilterJSS.val=""; RefreshUI(); });
                parentBrowser.ApplyHelp(clearTagFilterButton, "Clear the Tag search field of user entered text.");
                if (tagFilterJSS.val == "") clearTagFilterButton.gameObject.SetActive(false);

                createNewTagButton = CreateUIButton(parentBoxGO, 80f, 20f, "Create Tag", 22, 155, -2, AnchorPresets.topLeft);
                createNewTagButton.buttonColor = Color.green;
                createNewTagButton.textColor = Color.black;
                RefreshNewTagButton();
                createNewTagButton.button.onClick.AddListener(delegate {
                    var newTag = TagManifest.CreateTag(tagNameFilter, TagCategory.userDefined, filtersResourceType);
                    resourceSelectionsUI.SetSelectionTag(filtersResourceType, filtersSelectionMode, newTag, TriState.active);
                    RefreshUI();
                });
                parentBrowser.ApplyHelp(createNewTagButton, "Create a new Tag and assign it to the selected resources.");

//                var tagCategorryFilterUID = CreatePopup(parentBoxGO, tagCategoryFilterJSEnum, 160f, 20f, "Tag Category", 20, tagSelectionBoxWidth - 160f, -2f,AnchorPresets.topLeft,true);

                tagSelectionsScrollableBoxGO = UI.CreateVScrollableContent(parentBoxGO, new Color(0.5f, 0.5f, 0.65f, 0.98f), AnchorPresets.topLeft, tagSelectionBoxWidth, 48f, new Vector2(2f, -24f));
                tagSelectionScrollableViewportContentGO = tagSelectionsScrollableBoxGO.transform.Find("scrollableContentViewportGO/vScrollableContentGO").gameObject;

                RePopulateTagSelectionsBox();
            }
            
        }

        private void RefreshNewTagButton()
        {
            if (tagNameFilter == "hide" || tagNameFilter == "favorite" || tagNameFilter.EndsWith(" ") || tagNameFilter =="" || TagManifest.GetTags(filtersResourceType).Select(x => x.tagName).Contains(tagNameFilter)) createNewTagButton.gameObject.SetActive(false);
            else createNewTagButton.gameObject.SetActive(true);
        }

        private void TagFilterNameCallback(string i)
        {
            tagNameFilter = i.ToLower();
            if (i.Length > 0)
            {
                char lastCharacter = tagNameFilter.Last();
                if (tagNameFilter.Length > 25 || tagNameFilter.EndsWith("  ") || (!char.IsLetterOrDigit(lastCharacter) && !tagNameFilter.EndsWith(" ")))
                {
                    string tagNameLessOneChar = i.Remove(i.Length - 1, 1);
                    tagNameFilter = tagNameLessOneChar.ToLower();
                    
                }
            }
            tagFilterJSS.val = tagNameFilter;
            tagFilterJSS.inputField.text = tagNameFilter;
            if (tagFilterJSS.val == "") clearTagFilterButton.gameObject.SetActive(false);
            else clearTagFilterButton.gameObject.SetActive(true);

            RefreshNewTagButton();
            RePopulateTagSelectionsBox();

        }
        private void ClearTagSelectionRows()
        {
            foreach (var rowGO in tagSelectionRows) GameObject.Destroy(rowGO);
            tagSelectionRows.Clear();
        }

        const float tagSelectionBoxWidth = 550f;
        private GameObject CreateSelectTagRow()
        {
            GameObject currentRowGO = UI.AddChildGOImage(tagSelectionScrollableViewportContentGO, new Color(0f, 0.8f, 0f, 0f), AnchorPresets.hStretchTop, tagSelectionBoxWidth, 20f, Vector2.zero);
            tagSelectionRows.Add(currentRowGO);

            return currentRowGO;
        }

        private float GetTagButtonWidth(Tag tag)
        {
            return Mathf.Clamp((tag.tagName.Length * 5)+25, 40f, 150f);

        }

        public void RePopulateTagSelectionsBox()
        {
            foreach (var toggleKVP in tagJSBtoTriStateToggle)
            {

                GameObject.Destroy(toggleKVP.Value.toggleUID.gameObject);
                uiElements.Remove(toggleKVP.Value.toggleUID);
            }

            ClearTagSelectionRows();

            tagJSBtoTriStateToggle.Clear();
            tagJSBtoTag.Clear();

            float currentRowTagsWidth = 0f;

            List<Tag> availableTags = TagManifest.GetTags(filtersResourceType).Where(t => t.tagName.StartsWith(tagNameFilter)).ToList();

            if (availableTags.Count == 0) return;

            GameObject currentRowGO = CreateSelectTagRow();
            foreach (var tag in availableTags)
            {
                float nextButtonWidth = GetTagButtonWidth(tag);

                if (currentRowTagsWidth + nextButtonWidth  > tagSelectionBoxWidth)
                {
                    currentRowGO = CreateSelectTagRow();
                    currentRowTagsWidth = 0f;
                }
                var tagJSB = new JSONStorableBool(tag.tagName, false, TagCallback);
                var tagToggle = CreateTriStateToggle(currentRowGO, tagJSB, nextButtonWidth, 20f, tag.tagName, 20, currentRowTagsWidth, 0, AnchorPresets.middleLeft);
                tagToggle.RefreshToggle(resourceSelectionsUI.IsSelectionTagged(filtersResourceType, filtersSelectionMode, tag));
                tagJSBtoTag.Add(tagJSB, tag);
                tagJSBtoTriStateToggle.Add(tagJSB, tagToggle);

                tagToggle.backgroundColor = TagCategory.GetColor(tag.tagCategory);

                TagPointerEnterExitHelp contextHelp = tagToggle.toggle.gameObject.AddComponent<TagPointerEnterExitHelp>();
                contextHelp.Init(tag, parentBrowser.helpTextJSS, resourceSelectionsUI.GetResourceTypeSelections(browserFiltersUI.resourceTypeJSEnum.val),browserFiltersUI);

                currentRowTagsWidth += nextButtonWidth;

                currentRowTagsWidth += 3f;
            }
        }

        public void TagCallback(JSONStorableBool valueJSB)
        {
            int state = resourceSelectionsUI.IsSelectionTagged(filtersResourceType, filtersSelectionMode, tagJSBtoTag[valueJSB]);

            int targetTriState = TriState.active;
            if (state == TriState.active) targetTriState = TriState.inactive;

            resourceSelectionsUI.SetSelectionTag(filtersResourceType, filtersSelectionMode, tagJSBtoTag[valueJSB], targetTriState);
            tagJSBtoTriStateToggle[valueJSB].RefreshToggle(targetTriState);

            if (parentBrowser.browserFiltersUI.tagsFilterJSMultiEnum.valType!=JSONStorableMultiEnumStringChooser.topEnumValFlag || parentBrowser.browserFiltersUI.tagsFilterJSMultiEnum.valTopEnum!=TagGroups.anyTags) parentBrowser.RefreshFilterResults();
        }

        public void HiddenCallback(bool value)
        {
            int state = resourceSelectionsUI.IsSelectionHidden(filtersResourceType, filtersSelectionMode);

            if (state == TriState.active)
            {
                resourceSelectionsUI.SetSelectionHidden(filtersResourceType, filtersSelectionMode, TriState.inactive);
                hideToggle.RefreshToggle(TriState.inactive);
            }
            else
            {
                resourceSelectionsUI.SetSelectionHidden(filtersResourceType, filtersSelectionMode, TriState.active);
                hideToggle.RefreshToggle(TriState.active);
            }

            if (!parentBrowser.browserFiltersUI.showHiddenJSB.val)parentBrowser.RefreshFilterResults();
            
        }

        public void FavouriteCallback(bool value)
        {
            int state = resourceSelectionsUI.IsSelectionFavourite(filtersResourceType, filtersSelectionMode);

            if (state == TriState.active)
            {
                resourceSelectionsUI.SetSelectionFavourite(filtersResourceType, filtersSelectionMode, TriState.inactive);
                favToggle.RefreshToggle(TriState.inactive);
            }
            else
            {
                resourceSelectionsUI.SetSelectionFavourite(filtersResourceType, filtersSelectionMode, TriState.active);
                favToggle.RefreshToggle(TriState.active);
            }

            if (parentBrowser.browserFiltersUI.onlyFavouritesJSB.val) parentBrowser.RefreshFilterResults();
        }

    }


}
